# -*- coding: UTF-8 -*-
"""Definitions for `Ensembler` class."""

import gc
import sys
import time

import numpy as np
import scipy
from astrocats.catalog.model import MODEL
from astrocats.catalog.quantity import QUANTITY
from emcee.autocorr import AutocorrError
from mosfit.mossampler import MOSSampler
from mosfit.samplers.sampler import Sampler
from mosfit.utils import calculate_WAIC, pretty_num


class Ensembler(Sampler):
    """Fit transient events with the provided model."""

    _MAX_ACORC = 5
    _REPLACE_AGE = 20

    def __init__(
        self, fitter, model=None, iterations=2000, burn=None, post_burn=None,
            num_temps=1, num_walkers=None, convergence_criteria=None,
            convergence_type='psrf', gibbs=False, fracking=True,
            frack_step=20, **kwargs):
        """Initialize `Ensembler` class."""
        super(Ensembler, self).__init__(
            fitter, num_walkers=num_walkers, **kwargs)

        self._model = model
        self._iterations = iterations
        self._burn = burn
        self._post_burn = post_burn
        self._num_temps = num_temps
        self._cc = convergence_criteria
        self._ct = convergence_type
        self._gibbs = gibbs
        self._fracking = fracking
        self._frack_step = frack_step

        self._upload_model = None
        self._WAIC = None

    def append_output(self, modeldict):
        """Append output from the ensembler to the model description."""
        self._WAIC = None
        if self._iterations > 0:
            self._WAIC = calculate_WAIC(self._scores)
            modeldict[MODEL.SCORE] = {
                QUANTITY.VALUE: str(self._WAIC),
                QUANTITY.KIND: 'WAIC'
            }
            modeldict[MODEL.CONVERGENCE] = []
            if self._psrf < np.inf:
                modeldict[MODEL.CONVERGENCE].append(
                    {
                        QUANTITY.VALUE: str(self._psrf),
                        QUANTITY.KIND: 'psrf'
                    }
                )
            if self._acor and self._aacort > 0:
                acortimes = '<' if self._aa < self._MAX_ACORC else ''
                acortimes += str(np.int(float(self._emi -
                                              self._ams) / self._actc))
                modeldict[MODEL.CONVERGENCE].append(
                    {
                        QUANTITY.VALUE: str(acortimes),
                        QUANTITY.KIND: 'autocorrelationtimes'
                    }
                )
            modeldict[MODEL.STEPS] = str(self._emi)

    def prepare_output(self, check_upload_quality, upload):
        """Prepare output for writing to disk and uploading."""
        prt = self._printer

        if check_upload_quality:
            if self._WAIC is None:
                self._upload_model = False
            elif self._WAIC is not None and self._WAIC < 0.0:
                if upload:
                    prt.message('no_ul_waic', ['' if self._WAIC is None
                                               else pretty_num(self._WAIC)])
                self._upload_model = False

        if len(self._all_chain):
            self._pout = self._all_chain[:, :, -1, :]
            self._lnprobout = self._all_lnprob[:, :, -1]
            self._lnlikeout = self._all_lnlike[:, :, -1]
        else:
            self._pout = self._p
            self._lnprobout = self._lnprob
            self._lnlikeout = self._lnlike

        weight = 1.0 / (self._nwalkers * self._ntemps)
        self._weights = np.full_like(self._lnlikeout, weight)

        # Here, we append to the vector of walkers from the full chain based
        # upon the value of acort (the autocorrelation timescale).
        if self._acor and self._aacort > 0 and self._aa == self._MAX_ACORC:
            actc0 = int(np.ceil(self._aacort))
            for i in range(1, np.int(float(self._emi - self._ams) / actc0)):
                self._pout = np.concatenate(
                    (self._all_chain[:, :, -i * self._actc, :], self._pout),
                    axis=1)
                self._lnprobout = np.concatenate(
                    (self._all_lnprob[:, :, -i * self._actc],
                     self._lnprobout), axis=1)
                self._lnlikeout = np.concatenate(
                    (self._all_lnlike[:, :, -i * self._actc],
                     self._lnlikeout), axis=1)
                self._weights = np.full_like(self._lnlikeout, weight)

    def run(self, walker_data):
        """Use ensemble sampling to determine posteriors."""
        from mosfit.fitter import draw_walker, frack, ln_likelihood, ln_prior

        prt = self._printer

        self._emcee_est_t = 0.0
        self._bh_est_t = 0.0
        if self._burn is not None:
            self._burn_in = min(self._burn, self._iterations)
        elif self._post_burn is not None:
            self._burn_in = max(self._iterations - self._post_burn, 0)
        else:
            self._burn_in = int(np.round(self._iterations / 2))

        self._ntemps, ndim = (
            self._num_temps, self._model._num_free_parameters)

        if self._num_walkers:
            self._nwalkers = self._num_walkers
        else:
            self._nwalkers = 2 * ndim

        test_walker = self._iterations > 0
        self._lnprob = None
        self._lnlike = None
        pool_size = max(self._pool.size, 1)
        # Derived so only half a walker redrawn with Gaussian distribution.
        redraw_mult = 0.5 * np.sqrt(
            2) * scipy.special.erfinv(float(
                self._nwalkers - 1) / self._nwalkers)

        prt.message('nmeas_nfree', [self._model._num_measurements, ndim])
        if test_walker:
            if self._model._num_measurements <= ndim:
                prt.message('too_few_walkers', warning=True)
            if self._nwalkers < 10 * ndim:
                prt.message('want_more_walkers', [10 * ndim, self._nwalkers],
                            warning=True)
        p0 = [[] for x in range(self._ntemps)]

        # Generate walker positions based upon loaded walker data, if
        # available.
        walkers_pool = []
        walker_weights = []
        nmodels = len(set([x[0] for x in walker_data]))
        wp_extra = 0
        while len(walkers_pool) < len(walker_data):
            appended_walker = False
            for walk in walker_data:
                if (len(walkers_pool) + wp_extra) % nmodels != walk[0]:
                    continue
                new_walk = np.full(self._model._num_free_parameters, None)
                for k, key in enumerate(self._model._free_parameters):
                    param = self._model._modules[key]
                    walk_param = walk[1].get(key)
                    if walk_param is None or 'value' not in walk_param:
                        continue
                    if param:
                        val = param.fraction(
                            walk_param['value'], self._iterations != 0)
                        if not np.isnan(val):
                            new_walk[k] = val
                walkers_pool.append(new_walk)
                walker_weights.append(walk[2])
                appended_walker = True
            if not appended_walker:
                wp_extra += 1

        # Make sure weights are normalized.
        if None not in walker_weights:
            totw = np.sum(walker_weights)
            walker_weights = [x / totw for x in walker_weights]

        # Draw walker positions. This is either done from the priors or from
        # loaded walker data. If some parameters are not available from the
        # loaded walker data they will be drawn from their priors instead.
        pool_len = len(walkers_pool)
        for i, pt in enumerate(p0):
            dwscores = []
            while len(p0[i]) < self._nwalkers:
                prt.status(
                    self,
                    desc='drawing_walkers',
                    iterations=[
                        i * self._nwalkers + len(p0[i]) + 1,
                        self._nwalkers * self._ntemps])

                if self._pool.size == 0 or pool_len:
                    self._p, score = draw_walker(
                        test_walker, walkers_pool,
                        replace=pool_len < self._ntemps * self._nwalkers,
                        weights=walker_weights)
                    p0[i].append(self._p)
                    dwscores.append(score)
                else:
                    nmap = min(self._nwalkers -
                               len(p0[i]), max(self._pool.size, 10))
                    dws = self._pool.map(draw_walker, [test_walker] * nmap)
                    p0[i].extend([x[0] for x in dws])
                    dwscores.extend([x[1] for x in dws])

                if self._fitter._draw_above_likelihood is not False:
                    self._fitter._draw_above_likelihood = np.mean(dwscores)

        prt.message('initial_draws', inline=True)
        self._p = list(p0)

        self._emi = 0
        self._acor = None
        self._aacort = -1
        self._aa = 0
        self._psrf = np.inf
        self._all_chain = np.array([])
        self._scores = np.ones((self._ntemps, self._nwalkers)) * -np.inf

        tft = 0.0  # Total self._fracking time
        sli = 1.0  # Keep track of how many times chain halved
        s_exception = None
        kmat = None
        ages = np.zeros((self._ntemps, self._nwalkers), dtype=int)
        oldp = self._p

        max_chunk = 1000
        kmat_chunk = 5
        iter_chunks = int(np.ceil(float(self._iterations) / max_chunk))
        iter_arr = [max_chunk if xi < iter_chunks - 1 else
                    self._iterations - max_chunk * (iter_chunks - 1)
                    for xi, x in enumerate(range(iter_chunks))]
        # Make sure a chunk separation is located at self._burn_in
        chunk_is = sorted(set(
            np.concatenate(([0, self._burn_in], np.cumsum(iter_arr)))))
        iter_arr = np.diff(chunk_is)

        # The argument of the for loop runs emcee, after each iteration of
        # emcee the contents of the for loop are executed.
        converged = False
        exceeded_walltime = False
        ici = 0

        try:
            if self._iterations > 0:
                sampler = MOSSampler(
                    self._ntemps, self._nwalkers, ndim, ln_likelihood,
                    ln_prior, pool=self._pool)
            while (self._iterations > 0 and (
                    self._cc is not None or ici < len(iter_arr))):
                slr = int(np.round(sli))
                ic = (max_chunk if self._cc is not None else
                      iter_arr[ici])
                if exceeded_walltime:
                    break
                if (self._cc is not None and converged and
                        self._emi > self._iterations):
                    break
                for li, (
                        self._p, self._lnprob, self._lnlike) in enumerate(
                            sampler.sample(
                                self._p, iterations=ic, gibbs=self._gibbs if
                                self._emi >= self._burn_in else True)):
                    if (self._fitter._maximum_walltime is not False and
                            self.time_running() >
                            self._fitter._maximum_walltime):
                        prt.message('exceeded_walltime', warning=True)
                        exceeded_walltime = True
                        break
                    self._emi = self._emi + 1
                    emim1 = self._emi - 1
                    messages = []

                    # Increment the age of each walker if their positions are
                    # unchanged.
                    for ti in range(self._ntemps):
                        for wi in range(self._nwalkers):
                            if np.array_equal(self._p[ti][wi], oldp[ti][wi]):
                                ages[ti][wi] += 1
                            else:
                                ages[ti][wi] = 0

                    # Record then reset sampler proposal/acceptance counts.
                    accepts = list(
                        np.mean(sampler.nprop_accepted / sampler.nprop,
                                axis=1))
                    sampler.nprop = np.zeros(
                        (sampler.ntemps, sampler.nwalkers), dtype=np.float)
                    sampler.nprop_accepted = np.zeros(
                        (sampler.ntemps, sampler.nwalkers),
                        dtype=np.float)

                    # During self._burn-in only, redraw any walkers with scores
                    # significantly worse than their peers, or those that are
                    # stale (i.e. remained in the same position for a long
                    # time).
                    if emim1 <= self._burn_in:
                        pmedian = [np.median(x) for x in self._lnprob]
                        pmead = [np.mean([abs(y - pmedian) for y in x])
                                 for x in self._lnprob]
                        redraw_count = 0
                        bad_redraws = 0
                        for ti, tprob in enumerate(self._lnprob):
                            for wi, wprob in enumerate(tprob):
                                if (wprob <= pmedian[ti] -
                                    max(redraw_mult * pmead[ti],
                                        float(self._nwalkers)) or
                                        np.isnan(wprob) or
                                        ages[ti][wi] >= self._REPLACE_AGE):
                                    redraw_count = redraw_count + 1
                                    dxx = np.random.normal(
                                        scale=0.01, size=ndim)
                                    tar_x = np.array(
                                        self._p[np.random.randint(
                                            self._ntemps)][
                                            np.random.randint(self._nwalkers)])
                                    # Reflect if out of bounds.
                                    new_x = np.clip(np.where(
                                        np.where(tar_x + dxx < 1.0,
                                                 tar_x + dxx,
                                                 tar_x - dxx) > 0.0,
                                        tar_x + dxx, tar_x - dxx), 0.0, 1.0)
                                    new_like = ln_likelihood(new_x)
                                    new_prob = new_like + ln_prior(new_x)
                                    if new_prob > wprob or np.isnan(wprob):
                                        self._p[ti][wi] = new_x
                                        self._lnlike[ti][wi] = new_like
                                        self._lnprob[ti][wi] = new_prob
                                    else:
                                        bad_redraws = bad_redraws + 1
                        if redraw_count > 0:
                            messages.append(
                                '{:.0%} redraw, {}/{} success'.format(
                                    redraw_count /
                                    (self._nwalkers * self._ntemps),
                                    redraw_count - bad_redraws, redraw_count))

                    oldp = self._p.copy()

                    # Calculate the autocorrelation time.
                    low = 10
                    asize = 0.5 * (emim1 - self._burn_in) / low
                    if asize >= 0 and self._ct == 'acor':
                        acorc = max(
                            1, min(self._MAX_ACORC,
                                   int(np.floor(0.5 * self._emi / low))))
                        self._aacort = -1.0
                        self._aa = 0
                        self._ams = self._burn_in
                        cur_chain = (np.concatenate(
                            (self._all_chain,
                             sampler.chain[:, :, :li + 1:slr, :]),
                            axis=2) if len(self._all_chain) else
                            sampler.chain[:, :, :li + 1:slr, :])
                        for a in range(acorc, 1, -1):
                            ms = self._burn_in
                            if ms >= self._emi - low:
                                break
                            try:
                                acorts = sampler.get_autocorr_time(
                                    chain=cur_chain, low=low, c=a,
                                    min_step=int(np.round(float(ms) / sli)),
                                    max_walkers=5, fast=True)
                                acort = max([
                                    max(x)
                                    for x in acorts
                                ])
                            except AutocorrError:
                                continue
                            else:
                                self._aa = a
                                self._aacort = acort * sli
                                self._ams = ms
                                break
                        self._acor = [self._aacort, self._aa, self._ams]

                        self._actc = int(np.ceil(self._aacort / sli))
                        actn = np.int(
                            float(self._emi - self._ams) / self._actc)

                        if (self._cc is not None and
                            actn >= self._cc and
                                self._emi > self._iterations):
                            prt.message('converged')
                            converged = True
                            break

                    # Calculate the PSRF (Gelman-Rubin statistic).
                    if li > 1 and self._emi > self._burn_in + 2:
                        cur_chain = (np.concatenate(
                            (self._all_chain,
                             sampler.chain[:, :, :li + 1:slr, :]),
                            axis=2) if len(self._all_chain) else
                            sampler.chain[:, :, :li + 1:slr, :])
                        vws = np.zeros((self._ntemps, ndim))
                        for ti in range(self._ntemps):
                            for xi in range(ndim):
                                vchain = cur_chain[
                                    ti, :, int(np.floor(
                                        self._burn_in / sli)):, xi]
                                vws[ti][xi] = self.psrf(vchain)
                        self._psrf = np.max(vws)
                        if np.isnan(self._psrf):
                            self._psrf = np.inf

                        if (self._ct == 'psrf' and
                            self._cc is not None and
                            self._psrf < self._cc and
                                self._emi > self._iterations):
                            prt.message('converged')
                            converged = True
                            break

                    if self._cc is not None:
                        self._emcee_est_t = -1.0
                    else:
                        self._emcee_est_t = float(
                            self.time_running() - tft) / self._emi * (
                            self._iterations - self._emi
                        ) + tft / self._emi * max(
                                0, self._burn_in - self._emi)

                    # Perform self._fracking if we are still in the self._burn
                    # in phase and iteration count is a multiple of the frack
                    # step.
                    frack_now = (self._fracking and self._frack_step != 0 and
                                 self._emi <= self._burn_in and
                                 self._emi % self._frack_step == 0)

                    self._scores = [np.array(x) for x in self._lnprob]
                    if emim1 % kmat_chunk == 0:
                        sout = self._model.run_stack(
                            self._p[np.unravel_index(
                                np.argmax(self._lnprob), self._lnprob.shape)],
                            root='objective')
                        kmat = sout.get('kmat')
                        kdiag = sout.get('kdiagonal')
                        variance = sout.get('obandvs', sout.get('variance'))
                        if kdiag is not None and kmat is not None:
                            kmat[np.diag_indices_from(kmat)] += kdiag
                        elif kdiag is not None and kmat is None:
                            kmat = np.diag(kdiag + variance)
                    prt.status(
                        self,
                        desc='fracking' if frack_now else
                        ('burning' if self._emi < self._burn_in
                         else 'walking'),
                        scores=self._scores,
                        kmat=kmat,
                        accepts=accepts,
                        iterations=[self._emi, None if
                                    self._cc is not None else
                                    self._iterations],
                        acor=self._acor,
                        psrf=[self._psrf, self._burn_in],
                        messages=messages,
                        make_space=emim1 == 0,
                        convergence_type=self._ct,
                        convergence_criteria=self._cc,
                        time_running=self.time_running(),
                        maximum_walltime=self._fitter._maximum_walltime)

                    if s_exception:
                        break

                    if not frack_now:
                        continue

                    # Fracking starts here
                    sft = time.time()
                    ijperms = [[x, y] for x in range(self._ntemps)
                               for y in range(self._nwalkers)]
                    ijprobs = np.array([
                        1.0
                        # self._lnprob[x][y]
                        for x in range(self._ntemps) for y in range(
                            self._nwalkers)
                    ])
                    ijprobs -= max(ijprobs)
                    ijprobs = [np.exp(0.1 * x) for x in ijprobs]
                    ijprobs /= sum([x for x in ijprobs if not np.isnan(x)])
                    nonzeros = len([x for x in ijprobs if x > 0.0])
                    selijs = [
                        ijperms[x]
                        for x in np.random.choice(
                            range(len(ijperms)),
                            pool_size,
                            p=ijprobs,
                            replace=(pool_size > nonzeros))
                    ]

                    bhwalkers = [self._p[i][j] for i, j in selijs]

                    seeds = [
                        int(round(time.time() * 1000.0)) % 4294900000 + x
                        for x in range(len(bhwalkers))
                    ]
                    frack_args = list(zip(bhwalkers, seeds))
                    bhs = list(self._pool.map(frack, frack_args))
                    for bhi, bh in enumerate(bhs):
                        (wi, ti) = tuple(selijs[bhi])
                        if -bh.fun > self._lnprob[wi][ti]:
                            self._p[wi][ti] = bh.x
                            like = ln_likelihood(bh.x)
                            self._lnprob[wi][ti] = like + ln_prior(bh.x)
                            self._lnlike[wi][ti] = like
                    self._scores = [[-x.fun for x in bhs]]
                    prt.status(
                        self,
                        desc='fracking_results',
                        scores=self._scores,
                        kmat=kmat,
                        fracking=True,
                        iterations=[self._emi, None if
                                    self._cc is not None else
                                    self._iterations],
                        convergence_type=self._ct,
                        convergence_criteria=self._cc,
                        time_running=self.time_running(),
                        maximum_walltime=self._fitter._maximum_walltime)
                    tft = tft + time.time() - sft
                    if s_exception:
                        break

                if ici == 0:
                    self._all_chain = sampler.chain[:, :, :li + 1:slr, :]
                    self._all_lnprob = sampler.lnprobability[:, :, :li + 1:slr]
                    self._all_lnlike = sampler.lnlikelihood[:, :, :li + 1:slr]
                else:
                    self._all_chain = np.concatenate(
                        (self._all_chain, sampler.chain[:, :, :li + 1:slr, :]),
                        axis=2)
                    self._all_lnprob = np.concatenate(
                        (self._all_lnprob,
                         sampler.lnprobability[:, :, :li + 1:slr]),
                        axis=2)
                    self._all_lnlike = np.concatenate(
                        (self._all_lnlike,
                         sampler.lnlikelihood[:, :, :li + 1:slr]),
                        axis=2)

                mem_mb = (self._all_chain.nbytes + self._all_lnprob.nbytes +
                          self._all_lnlike.nbytes) / (1024. * 1024.)

                if self._fitter._debug:
                    prt.prt('Memory `{}`'.format(mem_mb), wrapped=True)

                if mem_mb > self._fitter._maximum_memory:
                    sfrac = float(
                        self._all_lnprob.shape[-1]) / self._all_lnprob[
                            :, :, ::2].shape[-1]
                    self._all_chain = self._all_chain[:, :, ::2, :]
                    self._all_lnprob = self._all_lnprob[:, :, ::2]
                    self._all_lnlike = self._all_lnlike[:, :, ::2]
                    sli *= sfrac
                    if self._fitter._debug:
                        prt.prt(
                            'Memory halved, sli: {}'.format(sli),
                            wrapped=True)

                sampler.reset()
                gc.collect()
                ici = ici + 1

        except (KeyboardInterrupt, SystemExit):
            prt.message('ctrl_c', error=True, prefix=False, color='!r')
            s_exception = sys.exc_info()
        except Exception:
            raise

        if s_exception is not None:
            self._pool.close()
            if (not prt.prompt('mc_interrupted')):
                sys.exit()

        msg_criteria = (
            1.1 if self._cc is None else self._cc)
        if (test_walker and self._ct == 'psrf' and
                msg_criteria is not None and self._psrf > msg_criteria):
            prt.message('not_converged', [
                'default' if self._cc is None else 'specified',
                msg_criteria], warning=True)



"""Definitions for the `LightCurve` class."""
from collections import OrderedDict

import numpy as np

from mosfit.modules.outputs.output import Output


# Important: Only define one ``Module`` class per file.


class LightCurve(Output):
    """Output a light curve to disk."""

    _lc_keys = [
        'magnitudes', 'e_magnitudes', 'model_observations', 'countrates',
        'e_countrates', 'all_telescopes', 'all_bands', 'all_systems',
        'all_instruments', 'all_bandsets', 'all_modes', 'all_times',
        'all_frequencies', 'observed', 'all_band_indices', 'observation_types'
    ]

    def __init__(self, **kwargs):
        """Initialize module."""
        super(LightCurve, self).__init__(**kwargs)
        self._dense_keys = self._lc_keys
        self._limiting_magnitude = self._model._fitter._limiting_magnitude

    def process(self, **kwargs):
        """Process module."""
        # First, rename some keys.
        output = OrderedDict()
        for key in sorted(kwargs.keys()):
            if key in self._dense_keys:
                continue
            output[key] = kwargs[key]
        for key in self._dense_keys:
            output[key.replace('all_', '')] = kwargs[key]

        if self._limiting_magnitude is not None:
            ls = 0.0
            if isinstance(self._limiting_magnitude, list):
                lm = float(self._limiting_magnitude[0])
                if len(self._limiting_magnitude) > 1:
                    ls = float(self._limiting_magnitude[1])
            else:
                lm = self._limiting_magnitude

            lmo = len(output['model_observations'])

            omags = output['observation_types'] == 'magnitude'
            output['model_variances'] = np.zeros_like(output[
                'model_observations'])
            output['model_upper_limits'] = np.full(lmo, False)
            lms = lm + ls * np.random.randn(lmo)
            varias = 10.0 ** (-lms / 2.5)
            mods = 10.0 ** (
                -np.array(output['model_observations'][omags]) / 2.5)
            output['model_observations'][omags] = -2.5 * np.log10(
                varias[omags] * np.random.randn(len(omags)) + mods)
            obsas = 10.0 ** (
                -np.array(output['model_observations']) / 2.5)
            output['model_variances'][omags] = np.abs(-output[
                'model_observations'][omags] - 2.5 * (
                    np.log10(varias[omags] + obsas)))
            ul_mask = omags & (obsas < 3.0 * varias)
            output['model_upper_limits'] = ul_mask
            output['model_observations'][ul_mask] = lms[ul_mask]
            output['model_variances'][ul_mask] = 2.5 * (
                np.log10(2.0 * varias[ul_mask]) - np.log10(varias[ul_mask]))
            return output

        # Then, apply GP predictions, if available.
        if (all([x in kwargs
                 for x in ['kmat', 'kfmat', 'koamat', 'kaomat']]) and not
            any([kwargs[x] is None
                 for x in ['kmat', 'kfmat', 'koamat', 'kaomat']])):
            kmat = kwargs['kmat'] + np.diag(kwargs['kdiagonal'])
            if np.linalg.cond(kmat) > 1e10:
                output['model_variances'] = np.full(
                    len(output['model_observations']), kwargs['variance'])
            else:
                ikmat = np.linalg.inv(kmat)
                kfmatd = np.diagonal(kwargs['kfmat'])
                koamat = kwargs['koamat']
                kaomat = kwargs['kaomat']
                output['model_variances'] = np.sqrt(
                    kfmatd - np.diagonal(np.matmul(np.matmul(
                        kaomat, ikmat), koamat)))
        else:
            output['model_variances'] = np.full(
                len(output['model_observations']), kwargs['abandvs'])

        return output



# -*- coding: UTF-8 -*-
"""Definitions for `Sampler` class."""

import numpy as np
import time


class Sampler(object):
    """Sample the posterior distribution of a model against an observation."""

    _MIN_WEIGHT = 1e-4

    def __init__(self, fitter, num_walkers=None, **kwargs):
        """Initialize `Sampler` class."""
        self._printer = kwargs.get('printer')
        self._fitter = fitter
        self._pool = self._fitter._pool
        self._printer = self._fitter._printer

        self._num_walkers = num_walkers

    def get_samples(self):
        """Return samples from ensembler."""
        samples = np.array([a for b in self._pout for a in b])
        if self._lnprobout is None:
            return samples, None, np.array([
                1.0 / len(samples) for x in samples])
        probs = np.array([a for b in self._lnprobout for a in b])
        weights = np.array([a for b in self._weights for a in b])

        min_weight = self._MIN_WEIGHT / len(samples)

        sel = weights > min_weight
        samples = samples[sel]
        probs = probs[sel]
        weights = weights[sel]

        wsis = np.argsort(weights)

        samples = samples[wsis]
        probs = probs[wsis]
        weights = weights[wsis]

        return samples, probs, weights

    def run(self):
        """Run the sampler."""
        pass

    def psrf(self, chain):
        """Calculate PSRF for a chain."""
        m = len(chain)
        n = len(chain[0])
        mom = np.mean(np.mean(chain, axis=1))
        b = n / float(m - 1) * np.sum(
            (np.mean(chain, axis=1) - mom) ** 2)
        w = np.mean(np.var(chain, axis=1, ddof=1))
        v = float(n - 1) / float(n) * w + (b / float(n))
        return np.sqrt(v / w)

    def time_running(self):
        return time.time() - self._fitter._start_time



"""Definitions for the `Utility` class."""
import numpy as np

from mosfit.modules.utilities.utility import Utility


# Important: Only define one ``Module`` class per file.


class Operator(Utility):
    """Template class for photosphere Modules."""

    ops = {
        '+': np.add,
        '-': np.subtract,
        '*': np.multiply,
        '/': np.divide
    }

    def set_attributes(self, task):
        """Set key replacement dictionary."""
        Utility.set_attributes(self, task)
        self._operands = task.get('operands', [])
        if not self._operands:
            raise ValueError('`Operator` must have at least one operand.')
        self._result = task.get('result', 'result')
        self._op = self.ops.get(task.get('operator', '+'), np.add)

    def process(self, **kwargs):
        """Process module."""
        ops = [
            'dense_' + x if self._wants_dense and
            not x.startswith('dense_') else x for x in self._operands]
        result = kwargs[ops[0]]
        for op in ops[1:]:
            result = self._op(result, kwargs[op])
        return {self.dense_key(self._result): result}



"""Definitions for the `AllTimes` class."""
from collections import OrderedDict

import numpy as np
from mosfit.modules.arrays.array import Array
from mosfit.utils import frequency_unit


# Important: Only define one ``Module`` class per file.


class AllTimes(Array):
    """Generate all times for which observations will be constructed.

    Create lists of observation times that associated with real observations
    and interpolations/extrapolations if such flags are passed to MOSFiT.
    """

    def __init__(self, **kwargs):
        """Initialize module."""
        super(AllTimes, self).__init__(**kwargs)
        self._obs_keys = [
            'times', 'bands', 'telescopes', 'systems', 'instruments', 'modes',
            'bandsets', 'frequencies', 'u_frequencies', 'zeropoints',
            'measures', 'observed']
        self._okeys = [i for i in self._obs_keys if i not in ['observed']]
        for key in self._obs_keys:
            setattr(self, '_' + key, [])

    def process(self, **kwargs):
        """Process module."""
        old_observations = tuple(
            getattr(self, '_' + x) for x in self._obs_keys)
        if (kwargs.get('root', 'output') == 'output' and
                'extra_times' in kwargs):
            obslist = (list(
                zip(*([kwargs.get(k) for k in self._okeys] +
                      [[True for x in range(len(kwargs['times']))]]))
            ) + list(
                zip(*([kwargs.get('extra_' + k) for k in self._okeys] +
                      [[False for x in range(len(kwargs['extra_times']))]]))))
            obslist.sort()

            self._all_observations = np.concatenate([
                np.atleast_2d(np.array(x, dtype=object))
                for x in obslist], axis=0).T
            for ki, key in enumerate(self._obs_keys):
                setattr(self, '_' + key, self._all_observations[ki])
        else:
            for key in list(
                    set(self._obs_keys) - set([
                        'frequencies', 'observed'])):
                setattr(self, '_' + key, kwargs[key])
            self._frequencies = np.array([
                x / frequency_unit(y) if x is not None else None
                for x, y in zip(kwargs['frequencies'], kwargs['u_frequencies'])
            ])
            self._observed = np.full_like(kwargs['times'], True, dtype=bool)
            self._all_observations = tuple(
                getattr(self, '_' + x) for x in self._obs_keys)

        outputs = OrderedDict(
            [('all_' + x, getattr(self, '_' + x))
             for x in list(set(self._obs_keys) - set(['observed']))])
        if any(not np.array_equal(x, y) for x, y in zip(
                old_observations, self._all_observations)):
            self._all_band_indices = np.array([
                (self._photometry.find_band_index(
                    b, telescope=t, instrument=i, mode=m, bandset=bs, system=s)
                 if f is None else -1)
                for ti, b, t, s, i, m, bs, f, uf, zps, msr, o
                in zip(*self._all_observations)
            ])
            self._observation_types = np.array([
                'magcount' if ('countrate' in msr and 'magnitude' in msr and
                               zp is not None and
                               self._fitter._prefer_fluxes) else
                self._photometry._band_kinds[bi] if bi >= 0 else
                'fluxdensity' for bi, zp, msr in zip(
                    self._all_band_indices, self._zeropoints, self._measures)
            ], dtype=object)
        outputs['all_band_indices'] = self._all_band_indices
        outputs['observation_types'] = self._observation_types
        outputs['observed'] = np.array(self._observed, dtype=bool)
        return outputs

    def receive_requests(self, **requests):
        """Receive requests from other ``Module`` objects."""
        self._photometry = requests.get('photometry', None)



"""Definitions for the `Simplefallback` class."""
from math import isnan

import numpy as np
from astrocats.catalog.source import SOURCE

from mosfit.constants import DAY_CGS
from mosfit.modules.engines.engine import Engine


# Important: Only define one ``Module`` class per file.


class Simplefallback(Engine):
    """Simple fallback energy input. Flat energy at first and then proportional to t**(-5/3)."""

    _REFERENCES = [
        {SOURCE.BIBCODE: '2013ApJ...772...30D'}
    ]

    def process(self, **kwargs):
        """Process module."""
        self._times = kwargs[self.key('dense_times')]
        self._L0 = kwargs[self.key('Lat1sec')]
        self._ton1 = kwargs[self.key('ton')]
        self._rest_t_explosion = kwargs[self.key('resttexplosion')]

        ts = [
            np.inf
            if self._rest_t_explosion > x else (x - self._rest_t_explosion)
            for x in self._times
        ]

        luminosities = [self._L0 / (t * DAY_CGS) ** (5./3.) if t - self._ton1 > 0 else self._L0 / (self._ton1 * DAY_CGS) ** (5./3.) for t in ts]
        luminosities = [0.0 if isnan(x) else x for x in luminosities]

        return {self.dense_key('luminosities'): luminosities}



"""Definitions for the `Blackbody` class."""
from math import pi

import numexpr as ne
import numpy as np
from astropy import constants as c
from astropy import units as u
from mosfit.constants import FOUR_PI
from mosfit.modules.seds.sed import SED


# Important: Only define one ``Module`` class per file.


class Blackbody(SED):
    """Blackbody spectral energy dist. for given temperature and radius."""

    C_CONST = c.c.cgs.value
    FLUX_CONST = FOUR_PI * (
        2.0 * c.h * c.c ** 2 * pi).cgs.value * u.Angstrom.cgs.scale
    X_CONST = (c.h * c.c / c.k_B).cgs.value
    STEF_CONST = (4.0 * pi * c.sigma_sb).cgs.value

    def process(self, **kwargs):
        """Process module."""
        lum_key = self.key('luminosities')
        kwargs = self.prepare_input(lum_key, **kwargs)
        self._luminosities = kwargs[lum_key]
        self._bands = kwargs['all_bands']
        self._band_indices = kwargs['all_band_indices']
        self._frequencies = kwargs['all_frequencies']
        self._radius_phot = kwargs[self.key('radiusphot')]
        self._temperature_phot = kwargs[self.key('temperaturephot')]
        xc = self.X_CONST  # noqa: F841
        fc = self.FLUX_CONST  # noqa: F841
        cc = self.C_CONST
        temperature_phot = self._temperature_phot

        # Some temp vars for speed.
        zp1 = 1.0 + kwargs[self.key('redshift')]
        Azp1 = u.Angstrom.cgs.scale / zp1
        czp1 = cc / zp1

        seds = []
        rest_wavs_dict = {}
        evaled = False

        for li, lum in enumerate(self._luminosities):
            bi = self._band_indices[li]
            if lum == 0.0:
                seds.append(np.zeros(len(
                    self._sample_wavelengths[bi]) if bi >= 0 else 1))
                continue

            if bi >= 0:
                rest_wavs = rest_wavs_dict.setdefault(
                    bi, self._sample_wavelengths[bi] * Azp1)
            else:
                rest_wavs = np.array(  # noqa: F841
                    [czp1 / self._frequencies[li]])

            radius_phot = self._radius_phot[li]  # noqa: F841
            temperature_phot = self._temperature_phot[li]  # noqa: F841

            if not evaled:
                seds.append(ne.evaluate(
                    'fc * radius_phot**2 / rest_wavs**5 / '
                    'expm1(xc / rest_wavs / temperature_phot)'))
                evaled = True
            else:
                seds.append(ne.re_evaluate())

            seds[-1][np.isnan(seds[-1])] = 0.0

        seds = self.add_to_existing_seds(seds, **kwargs)

        # Units of `seds` is ergs / s / Angstrom.
        return {'sample_wavelengths': self._sample_wavelengths, 'seds': seds}



"""Definitions for the `Kernel` class."""
from collections import OrderedDict

import numpy as np
from six import string_types

from mosfit.constants import ANG_CGS, C_CGS
from mosfit.modules.arrays.array import Array


# Important: Only define one ``Module`` class per file.


class Kernel(Array):
    """Calculate the kernel for use in computing the likelihood score."""

    MIN_COV_TERM = 1.0e-30

    def __init__(self, **kwargs):
        """Initialize module."""
        super(Kernel, self).__init__(**kwargs)
        self._times = np.array([])
        self._codeltatime = -1
        self._codeltalambda = -1
        self._type = kwargs.get('type', False)

    def process(self, **kwargs):
        """Process module."""
        self.preprocess(**kwargs)

        ret = OrderedDict()

        # If we are trying to krig between observations, we need an array with
        # dimensions equal to the number of intermediate observations.
        if self._type == 'full':
            kskey = 'kfmat'
        elif self._type == 'oa':
            kskey = 'koamat'
        elif self._type == 'ao':
            kskey = 'kaomat'
        else:
            kskey = 'kmat'

        # Get band variances
        self._variance = kwargs.get(self.key('variance'), 0.0)

        # Get array of real observations.
        self._observations = np.array([
            ct if (t == 'countrate' or t == 'magcount') else y if (
                t == 'magnitude') else fd if t == 'fluxdensity' else None
            for y, ct, fd, t in zip(
                self._mags, self._cts, self._fds, self._o_otypes)
        ])

        # Get array of model observations.
        self._model_observations = kwargs.get('model_observations', [])

        # Handle band-specific variances if that option is enabled.
        self._band_v_vars = OrderedDict()
        for key in kwargs:
            if key.startswith('variance-band-'):
                self._band_v_vars[key.split('-')[-1]] = kwargs[key]

        if self._variance_bands:
            self._o_variance_bands = [
                self._variance_bands[i] for i in self._all_band_indices]

            self._band_vs = np.array([
                self._band_v_vars.get(i, self._variance) if
                isinstance(i, string_types)
                else (i[0] * self._band_v_vars.get(i[1][0], self._variance) +
                      (1.0 - i[0]) * self._band_v_vars.get(
                          i[1][0], self._variance))
                for i in self._o_variance_bands])
        else:
            self._band_vs = np.full(
                len(self._all_band_indices), self._variance)

        # Compute relative errors for count-based observations.
        self._band_vs[self._count_inds] = (
            10.0 ** (self._band_vs[self._count_inds] / 2.5) - 1.0
            ) * self._model_observations[self._count_inds]

        self._o_band_vs = self._band_vs[self._observed]

        if self._type == 'full':
            self._band_vs_1 = self._band_vs
            self._band_vs_2 = self._band_vs
        elif self._type == 'oa':
            self._band_vs_1 = self._o_band_vs
            self._band_vs_2 = self._band_vs
        elif self._type == 'ao':
            self._band_vs_1 = self._band_vs
            self._band_vs_2 = self._o_band_vs
        else:
            self._band_vs_1 = self._o_band_vs
            self._band_vs_2 = self._o_band_vs

        if self._codeltatime >= 0 or self._codeltalambda >= 0:
            kmat = np.outer(self._band_vs_1, self._band_vs_2)

            if self._codeltatime >= 0:
                kmat *= np.exp(self._dt2mat / self._codeltatime ** 2)

            if self._codeltalambda >= 0:
                kmat *= np.exp(self._dl2mat / self._codeltalambda ** 2)

            ret[kskey] = kmat
        else:
            ret['abandvs'] = self._band_vs
            ret['obandvs'] = self._o_band_vs

        return ret

    def receive_requests(self, **requests):
        """Receive requests from other ``Module`` objects."""
        self._average_wavelengths = requests.get('average_wavelengths', [])
        self._variance_bands = requests.get('variance_bands', [])

    def preprocess(self, **kwargs):
        """Construct kernel distance arrays."""
        new_times = np.array(kwargs.get('all_times', []), dtype=float)
        self._codeltatime = kwargs.get(self.key('codeltatime'), -1)
        self._codeltalambda = kwargs.get(self.key('codeltalambda'), -1)
        if np.array_equiv(new_times, self._times) and self._preprocessed:
            return
        self._times = new_times
        self._all_band_indices = kwargs.get('all_band_indices', [])
        self._are_bands = np.array(self._all_band_indices) >= 0
        self._freqs = kwargs.get('all_frequencies', [])
        self._mags = np.array(kwargs.get('magnitudes', []))
        self._fds = np.array(kwargs.get('fluxdensities', []))
        self._cts = np.array(kwargs.get('countrates', []))
        self._u_freqs = kwargs.get('all_u_frequencies', [])
        self._waves = np.array([
            self._average_wavelengths[bi] if bi >= 0 else
            C_CGS / self._freqs[i] / ANG_CGS for i, bi in
            enumerate(self._all_band_indices)])
        self._observed = np.array(kwargs.get('observed', []), dtype=bool)
        self._observation_types = kwargs.get('observation_types')
        self._n_obs = len(self._observed)
        self._count_inds = self._observation_types != 'magnitude'

        self._o_times = self._times[self._observed]
        self._o_waves = self._waves[self._observed]
        self._o_otypes = self._observation_types[self._observed]

        if self._type == 'full':
            self._times_1 = self._times
            self._times_2 = self._times
            self._waves_1 = self._waves
            self._waves_2 = self._waves
        elif self._type == 'oa':
            self._times_1 = self._o_times
            self._times_2 = self._times
            self._waves_1 = self._o_waves
            self._waves_2 = self._waves
        elif self._type == 'ao':
            self._times_1 = self._times
            self._times_2 = self._o_times
            self._waves_1 = self._waves
            self._waves_2 = self._o_waves
        else:
            self._times_1 = self._o_times
            self._times_2 = self._o_times
            self._waves_1 = self._o_waves
            self._waves_2 = self._o_waves

        # Time deltas (radial distance) for covariance matrix.
        if self._codeltatime >= 0:
            self._dt2mat = self._times_1[:, None] - self._times_2[None, :]
            self._dt2mat **= 2
            self._dt2mat *= -0.5

        # Wavelength deltas (radial distance) for covariance matrix.
        if self._codeltalambda >= 0:
            self._dl2mat = self._waves_1[:, None] - self._waves_2[None, :]
            self._dl2mat **= 2
            self._dl2mat *= -0.5

        self._preprocessed = True



# -*- coding: UTF-8 -*-
"""A class for converting ASCII inputs to JSON."""
import csv
import os
import re
from collections import Counter, OrderedDict
from decimal import Decimal
from itertools import chain, product

import inflect
import numpy as np
from astrocats.catalog.catalog import Catalog
from astrocats.catalog.entry import ENTRY, Entry
from astrocats.catalog.key import KEY_TYPES, Key
from astrocats.catalog.photometry import (PHOTOMETRY, set_pd_mag_from_counts,
                                          set_pd_mag_from_flux_density)
from astrocats.catalog.quantity import QUANTITY
from astrocats.catalog.source import SOURCE
from astrocats.catalog.utils import jd_to_mjd
from astropy.io.ascii import Cds, Latex, read
from astropy.time import Time as astrotime
from mosfit.constants import KS_DAYS
from mosfit.utils import (entabbed_json_dump, get_mosfit_hash, is_bibcode,
                          is_date, is_datum, is_number, listify, name_clean,
                          replace_multiple)
from six import string_types


class Converter(object):
    """Convert ASCII formats to Open Catalog JSON schemas."""

    _MONTH_IDS = OrderedDict((
        ('January', '01'),
        ('February', '02'),
        ('March', '03'),
        ('April', '04'),
        ('June', '06'),
        ('July', '07'),
        ('August', '08'),
        ('September', '09'),
        ('October', '10'),
        ('November', '11'),
        ('December', '12'),
        ('Jan', '01'),
        ('Feb', '02'),
        ('Mar', '03'),
        ('Apr', '04'),
        ('May', '05'),
        ('Jun', '06'),
        ('Jul', '07'),
        ('Aug', '08'),
        ('Sep', '09'),
        ('Oct', '10'),
        ('Nov', '11'),
        ('Dec', '12')
    ))

    _DEFAULT_SOURCE = 'MOSFiT paper'

    _TRUE_VALS = ['t', 'true', 'T', 'True', '1', 'y', 'Y']
    _FALSE_VALS = ['f', 'false', 'F', 'False', '0', 'n', 'N']
    _EMPTY_VALS = ['nodata']

    def __init__(self, printer, require_source=False, **kwargs):
        """Initialize."""
        import pickle

        self._path = os.path.dirname(os.path.realpath(__file__))
        self._inflect = inflect.engine()
        self._printer = printer
        self._require_source = require_source

        self._estrs = [
            'err', '_err', 'err_', 'ERR', 'e_', '_e', '(err)', 'error',
            'uncertainty', 'sigma']
        self._emagstrs = self._estrs + [
            'magnitude error', 'e mag', 'e magnitude', 'dmag',
            'mag err', 'magerr', 'mag error']
        self._ecntstrs = self._estrs + [
            'flux error', 'e flux', 'e counts', 'count err', 'flux err',
            'countrate error', 'countrate err', 'e_flux']
        self._lmagstrs = ['l_']
        self._band_names = [
            'U', 'B', 'V', 'R', 'I', 'J', 'H', 'K', 'K_s', "Ks", "K'", 'u',
            'g', 'r', 'i', 'z', 'y', 'W1', 'W2', 'M2', "u'", "g'", "r'", "i'",
            "z'", 'C', 'Y', 'I1', 'I2', 'I3', 'I4', 'Open'
        ]
        ebands = [a + b for a, b in chain(
            product(self._ecntstrs, self._band_names),
            product(self._band_names, self._estrs))]
        lbands = [a + b for a, b in chain(
            product(self._lmagstrs, self._band_names))]
        self._emagstrs += ebands
        self._ecntstrs += ebands
        self._lmagstrs += lbands
        key_cache_path = os.path.join(
            self._path, 'cache', 'key_cache_{}.pickle'.format(
                get_mosfit_hash()))
        hks_loaded = False
        if os.path.isfile(key_cache_path):
            try:
                self._header_keys = pickle.load(open(key_cache_path, 'rb'))
                hks_loaded = True
            except Exception:
                printer.message('bad_header_pickle', warning=True)
                if hasattr(self, '_header_keys'):
                    del(self._header_keys)
        if not hks_loaded:
            self._header_keys = OrderedDict((
                (PHOTOMETRY.TIME, [
                    'time', 'mjd', ('jd', 'jd'), ('julian date', 'jd'),
                    ('date', 'yyyy-mm-dd'), 'day', (
                        'kiloseconds', 'kiloseconds')]),
                (PHOTOMETRY.SYSTEM, ['system', 'magsys', 'magnitude system']),
                (PHOTOMETRY.MAGNITUDE, [
                 'vega mag', 'ab mag', 'mag', 'magnitude']),
                (PHOTOMETRY.E_MAGNITUDE, self._emagstrs),
                (PHOTOMETRY.UPPER_LIMIT, self._lmagstrs),
                (PHOTOMETRY.TELESCOPE, ['tel', 'telescope']),
                (PHOTOMETRY.INSTRUMENT, ['inst', 'instrument']),
                (PHOTOMETRY.OBSERVER, ['observer']),
                (PHOTOMETRY.OBSERVATORY, ['observatory']),
                (PHOTOMETRY.BAND, ['passband', 'band', 'filter', 'filt',
                                   'flt']),
                (PHOTOMETRY.E_LOWER_MAGNITUDE, [a + ' ' + b for a, b in chain(
                    product(self._emagstrs, ['minus', 'lower']),
                    product(['minus', 'lower'], self._emagstrs))]),
                (PHOTOMETRY.E_UPPER_MAGNITUDE, [a + ' ' + b for a, b in chain(
                    product(self._emagstrs, ['plus', 'upper']),
                    product(['plus', 'upper'], self._emagstrs))]),
                (PHOTOMETRY.UPPER_LIMIT, [
                    'upper limit', 'upperlimit', 'l_mag', 'limit']),
                (PHOTOMETRY.COUNT_RATE, [
                    'count', 'counts', 'flux', 'count rate']),
                (PHOTOMETRY.E_COUNT_RATE, self._ecntstrs),
                (PHOTOMETRY.FLUX_DENSITY, ['flux density', 'fd', 'f_nu']),
                (PHOTOMETRY.E_FLUX_DENSITY, [
                    'e_flux_density', 'flux density error', 'e_fd',
                    'sigma_nu']),
                (PHOTOMETRY.U_FLUX_DENSITY, []),
                (PHOTOMETRY.ZERO_POINT, ['zero point', 'zp']),
                ('reference', ['reference', 'bibcode', 'source', 'origin']),
                (ENTRY.NAME, [
                    'event', 'transient', 'name', 'supernova', 'sne', 'id',
                    'identifier', 'object']),
                (ENTRY.REDSHIFT, ['redshift']),
                (ENTRY.HOST, ['host']),
                (ENTRY.LUM_DIST, [
                    'lumdist', 'luminosity distance', 'distance']),
                (ENTRY.COMOVING_DIST, ['comoving distance']),
                (ENTRY.RA, ['ra', 'right ascension', 'right_ascension']),
                (ENTRY.DEC, ['dec', 'declination']),
                (ENTRY.EBV, ['ebv', 'extinction']),
                # At the moment transient-specific keys are not in astrocats.
                (Key('claimedtype', KEY_TYPES.STRING), [
                    'claimedtype', 'type', 'claimed_type', 'claimed type'])
            ))

        self._critical_keys = [
            PHOTOMETRY.TIME, PHOTOMETRY.MAGNITUDE, PHOTOMETRY.COUNT_RATE,
            PHOTOMETRY.FLUX_DENSITY,
            PHOTOMETRY.BAND, PHOTOMETRY.E_COUNT_RATE,
            PHOTOMETRY.E_FLUX_DENSITY, PHOTOMETRY.ZERO_POINT]
        self._helpful_keys = [
            PHOTOMETRY.E_MAGNITUDE, PHOTOMETRY.INSTRUMENT,
            PHOTOMETRY.TELESCOPE]
        self._optional_keys = [
            PHOTOMETRY.ZERO_POINT, PHOTOMETRY.E_MAGNITUDE,
            PHOTOMETRY.U_FLUX_DENSITY]
        self._mc_keys = [
            PHOTOMETRY.MAGNITUDE, PHOTOMETRY.COUNT_RATE,
            PHOTOMETRY.FLUX_DENSITY]
        self._dep_keys = [
            PHOTOMETRY.E_MAGNITUDE, PHOTOMETRY.E_COUNT_RATE,
            PHOTOMETRY.E_FLUX_DENSITY, PHOTOMETRY.U_FLUX_DENSITY,
            PHOTOMETRY.BAND]
        self._purge_non_numeric_keys = [
            PHOTOMETRY.E_MAGNITUDE, PHOTOMETRY.E_LOWER_MAGNITUDE,
            PHOTOMETRY.E_UPPER_MAGNITUDE, PHOTOMETRY.E_COUNT_RATE,
            PHOTOMETRY.E_LOWER_COUNT_RATE, PHOTOMETRY.E_UPPER_COUNT_RATE,
            PHOTOMETRY.E_FLUX, PHOTOMETRY.E_LOWER_FLUX,
            PHOTOMETRY.E_UPPER_FLUX, PHOTOMETRY.E_UNABSORBED_FLUX,
            PHOTOMETRY.E_LOWER_UNABSORBED_FLUX,
            PHOTOMETRY.E_UPPER_UNABSORBED_FLUX]
        self._positive_keys = [
            PHOTOMETRY.MAGNITUDE] + self._purge_non_numeric_keys
        self._bool_keys = [PHOTOMETRY.UPPER_LIMIT]
        self._specify_keys = [
            PHOTOMETRY.BAND, PHOTOMETRY.INSTRUMENT, PHOTOMETRY.TELESCOPE]
        self._entry_keys = [
            ENTRY.COMOVING_DIST, ENTRY.REDSHIFT, ENTRY.LUM_DIST,
            ENTRY.RA, ENTRY.DEC, ENTRY.EBV, ENTRY.HOST, Key(
                'claimedtype', KEY_TYPES.STRING)]
        self._use_mc = False
        self._month_rep = re.compile(
            r'\b(' + '|'.join(self._MONTH_IDS.keys()) + r')\b')
        self._converted = []

        if not hks_loaded:
            for key in self._header_keys.keys():
                for val in self._header_keys[key]:
                    for i in range(val.count(' ')):
                        rep = val.replace(' ', '_', i + 1)
                        if rep not in self._header_keys[key]:
                            self._header_keys[key].append(rep)
                    for i in range(val.count(' ')):
                        rep = val.replace(' ', '', i + 1)
                        if rep not in self._header_keys[key]:
                            self._header_keys[key].append(rep)
            pickle.dump(self._header_keys, open(key_cache_path, 'wb'))

    def generate_event_list(self, event_list):
        """Generate a list of events and/or convert events to JSON format."""
        prt = self._printer
        cidict = OrderedDict()
        intro_shown = False
        check_all_files = None
        shared_sources = []

        new_event_list = []
        previous_file = None
        for event in event_list:
            rsource = {SOURCE.NAME: self._DEFAULT_SOURCE}
            use_self_source = None
            new_events = []
            toffset = Decimal('0')
            if ('.' in event and os.path.isfile(event) and
                    not event.endswith('.json')):
                with open(event, 'r') as f:
                    ftxt = f.read()

                # Try a couple of table formats from astropy.
                table = None
                try:
                    table = read(ftxt, Reader=Cds, guess=False)
                except Exception:
                    pass
                else:
                    prt.message('convert_cds')
                    flines = [table.colnames] + [
                        list(x) for x in np.array(table).tolist()]
                    for i in range(len(flines)):
                        flines[i] = [str(x) for x in flines[i]]

                try:
                    table = read(ftxt, Reader=Latex, guess=False)
                except Exception:
                    pass
                else:
                    prt.message('convert_latex')
                    flines = [table.colnames] + [
                        list(x) for x in np.array(table).tolist()]

                if table is None:
                    # Count to try and determine delimiter.
                    delims = [' ', '\t', ',', ';', '|', '&']
                    delimnames = [
                        'Space: ` `', 'Tab: `\t`', 'Comma: `,`',
                        'Semi-colon: `;`', 'Bar: `|`', 'Ampersand: `&`']
                    delim = None
                    delimcounts = [re.sub(
                        re.escape(y) + '+', y, re.sub(
                            ' ?[' + ''.join([re.escape(
                                x) for x in delims if x != y]) + ']' +
                            ' ?', '', ftxt)).count(
                                y) for y in delims]
                    maxdelimcount = max(delimcounts)
                    # Make sure at least one delimeter per line.
                    maxdelimavg = delimcounts[delimcounts.index(
                        maxdelimcount)] / len(ftxt.splitlines())
                    if maxdelimavg >= 1.0:
                        delim = delims[delimcounts.index(maxdelimcount)]
                    # If two delimiter options are close in count, ask user.
                    for i, x in enumerate(delimcounts):
                        if x > 0.5 * maxdelimcount and delims[i] != delim:
                            delim = None
                    if delim is None and maxdelimavg >= 1.0:
                        odelims = list(np.array(delimnames)[
                            np.array(delimcounts) > 0])
                        dchoice = prt.prompt(
                            'delim', kind='option', options=odelims,
                            none_string=prt.text('no_delimiter'))
                        if is_number(dchoice):
                            delim = delims[dchoice - 1]
                    if delim is not None:
                        ad = list(delims)
                        ad.remove(delim)
                        ad = ''.join(ad)

                    fsplit = ftxt.splitlines()

                    # If none of the rows contain numeric data, the file
                    # is likely a list of transient names.
                    flines = list(fsplit)

                    if (len(flines) and
                        (not any(any([
                            is_datum(x.strip()) or x == '' for x in (
                                y.split(delim) if delim is not None else
                                listify(y))]) for y in flines) or
                         len(flines) == 1)):
                        new_events = [
                            it.strip() for s in flines for it in (
                                s.split(delim) if delim is not None else
                                listify(s))]
                        new_event_list.extend(new_events)
                        continue

                    if delim is None:
                        raise ValueError(prt.text('delimiter_not_found'))

                if not intro_shown:
                    prt.message('converter_info')
                    intro_shown = True

                prt.message('converting_to_json', [event])

                if table is None:
                    # See if we need to append blank errors to upper limits.
                    tsplit = [
                        replace_multiple(x, ['$', '\\pm', '±', '-or+'], delim)
                        .strip(ad + '()# ').replace('′', "'")
                        for x in fsplit]

                    append_missing_errs = False
                    for fl in tsplit:
                        dfl = list(csv.reader([fl], delimiter=delim))[0]
                        if any([is_number(x.strip('(<>≤≥'))
                                for x in dfl]) and any([
                                any([y in x for y in [
                                    '(', '<', '>', '≥', '≤']])
                                for x in dfl]):
                            append_missing_errs = True
                            break

                    fsplit = [
                        replace_multiple(x, ['$', '\\pm', '±', '-or+'], delim)
                        .replace('(', delim + '(')
                        .strip(ad + '()# ').replace('′', "'")
                        for x in fsplit]
                    flines = []
                    for fs in fsplit:
                        # Replace repeated spaces if fixed-width
                        if delim in [' ']:
                            fsn = re.sub(r'(\s)\1+', r'\1', fs)
                        else:
                            fsn = fs
                        flines.append(list(
                            csv.reader([fsn], delimiter=delim))[0])

                    flines = [[
                        x.strip(ad + '#$()\\')
                        for x in y] for y in flines]

                    # Find band columns if they exist and insert error columns
                    # if they don't exist.
                    for fi, fl in enumerate(list(flines)):
                        flcopy = list(fl)
                        offset = 0
                        if not any([is_datum(x) for x in fl]):
                            for fci, fc in enumerate(fl):
                                if (fc in self._band_names and
                                    (fci == len(fl) - 1 or
                                     fl[fci + 1] not in self._emagstrs)):
                                    flcopy.insert(fci + 1 + offset, 'e_' + fc)
                                    offset += 1
                        flines[fi] = flcopy

                    # Append blank errors to upper limits.
                    if append_missing_errs:
                        # Find the most frequent column count. These are
                        # probably the tables we wish to read.
                        flens = [len(x) for x in flines]
                        ncols = Counter(flens).most_common(1)[0][0]

                        flines = [[x for y in [
                            ([z, '-'] if (any([w in z for w in [
                                '<', '>', '≤', '≥']]) or
                                z in self._EMPTY_VALS) else [z])
                            for z in fl] for x in y] if len(
                                fl) != ncols else fl for fl in flines]

                    # Find the most frequent column count. These are probably
                    # the tables we wish to read.
                    flens = [len(x) for x in flines]
                    ncols = Counter(flens).most_common(1)[0][0]

                    newlines = []
                    potential_name = None
                    for fi, fl in enumerate(flines):
                        if (len(fl) and flens[fi] == 1 and
                            fi < len(flines) - 1 and
                                flens[fi + 1] == ncols and not len(newlines)):
                            potential_name = fl[0]
                        if flens[fi] == ncols:
                            if potential_name is not None and any(
                                    [is_datum(x) for x in fl]):
                                newlines.append([potential_name] + list(fl))
                            else:
                                newlines.append(list(fl))
                    flines = newlines
                    for fi, fl in enumerate(flines):
                        if len(fl) == ncols and potential_name is not None:
                            if not any([is_datum(x) for x in fl]):
                                flines[fi] = ['name'] + list(fl)

                # If last row is numeric, then likely this is a file with
                # transient data.
                if (len(flines) > 1 and
                        any([is_datum(x) for x in flines[-1]])):

                    # Check that each row has the same number of columns.
                    if len(set([len(x) for x in flines])) > 1:
                        print(set([len(x) for x in flines]))
                        raise ValueError(
                            'Number of columns in each row not '
                            'consistent!')

                    if len(cidict) and len(
                            new_event_list) and check_all_files is None:
                        reps = [previous_file] if previous_file else [''.join(
                            new_event_list[-1].split('.')[:-1])]
                        check_all_files = not prt.prompt(
                            'check_all_files', reps=reps, kind='bool')

                    if check_all_files or check_all_files is None:
                        shared_sources = []

                    if len(cidict) and len(new_event_list) and check_all_files:
                        msg = ('is_file_same' if
                               previous_file else 'is_event_same')
                        reps = [previous_file] if previous_file else [''.join(
                            new_event_list[-1].split('.')[:-1])]
                        is_same = prt.prompt(msg, reps=reps, kind='bool')
                        if not is_same:
                            cidict = OrderedDict()

                    # If the first row has no numbers it is likely a header.
                    if not len(cidict):
                        self.assign_columns(cidict, flines)

                    perms = 1
                    for key in cidict:
                        if isinstance(cidict[key], list) and not isinstance(
                                cidict[key], string_types):
                            if cidict[key][0] != 'j':
                                perms = len(cidict[key])

                    # Get event name (if single event) or list of names from
                    # table.
                    event_names = []
                    if ENTRY.NAME in cidict:
                        for fi, fl in enumerate(flines):
                            flines[fi][cidict[ENTRY.NAME]] = name_clean(
                                fl[cidict[ENTRY.NAME]])
                        event_names = list(sorted(set([
                            x[cidict[ENTRY.NAME]] for x in flines[
                                self._first_data:]])))
                        new_events = [x + '.json' for x in event_names]
                    else:
                        new_event_name = '.'.join(event.split(
                            '.')[:-1]).split('/')[-1]
                        if check_all_files or check_all_files is None:
                            is_name = prt.prompt(
                                'is_event_name', reps=[new_event_name],
                                kind='bool', default='y')
                            if not is_name:
                                new_event_name = ''
                                while new_event_name.strip() == '':
                                    new_event_name = prt.prompt(
                                        'enter_name', kind='string')
                        event_names.append(new_event_name)
                        new_events = [new_event_name + '.json']

                    # Create a new event, populate the photometry, and dump
                    # to a JSON file in the run directory.
                    entries = OrderedDict([(x, Entry(name=x))
                                           for x in event_names])

                    # Clean up the data a bit now that we know the column
                    # identities.

                    # Strip common prefixes/suffixes from band names
                    if PHOTOMETRY.BAND in cidict:
                        bi = cidict[PHOTOMETRY.BAND]
                        for d in [True, False]:
                            if not isinstance(bi, (int, np.integer)):
                                break
                            strip_cols = []
                            lens = [len(x[bi])
                                    for x in flines[self._first_data:]]
                            llen = min(lens)
                            ra = range(llen) if d else range(-1, -llen - 1, -1)
                            for li in ra:
                                letter = None
                                for row in list(flines[self._first_data:]):
                                    if letter is None:
                                        letter = row[bi][li]
                                    elif row[bi][li] != letter:
                                        letter = None
                                        break
                                if letter is not None:
                                    strip_cols.append(li)
                                else:
                                    break
                            if len(strip_cols) == llen:
                                break
                            for ri in range(len(flines[self._first_data:])):
                                flines[self._first_data + ri][bi] = ''.join(
                                    [c for i, c in enumerate(flines[
                                        self._first_data + ri][bi])
                                     if (i if d else i - len(flines[
                                         self._first_data + ri][bi])) not in
                                     strip_cols])

                    if (PHOTOMETRY.TIME in cidict and
                            (not isinstance(cidict[PHOTOMETRY.TIME], list) or
                             len(cidict[PHOTOMETRY.TIME]) <= 2)):
                        bi = cidict[PHOTOMETRY.TIME]

                        bistr = bi[0].upper() if isinstance(
                            bi, list) and not isinstance(
                                bi, string_types) else 'MJD'

                        if isinstance(bi, list) and not isinstance(
                            bi, string_types) and isinstance(
                                bi[0], string_types):
                            bi = bi[-1]

                        mmtimes = None
                        try:
                            mmtimes = [float(x[bi])
                                       for x in flines[self._first_data:]]
                            mintime, maxtime = min(mmtimes), max(mmtimes)
                        except Exception:
                            pass

                        if not mmtimes:
                            pass
                        elif (bistr == 'MJD' and mintime < 10000 or
                              bistr == 'JD' and mintime < 2410000 or
                                bistr in ['KILOSECONDS', 'SECONDS']):
                            while True:
                                pstr = ('s_offset' if bistr in [
                                    'KILOSECONDS', 'SECONDS'] else
                                    'small_time_offset')
                                try:
                                    response = prt.prompt(pstr, [
                                        bistr for x in range(3)],
                                        kind='string')
                                    if response is not None:
                                        toffset = Decimal(response)
                                    break
                                except Exception:
                                    pass
                        elif maxtime > 90000 and bistr != 'JD':
                            isjd = prt.prompt(
                                'large_time_offset',
                                kind='bool', default='y')
                            if isjd:
                                toffset = Decimal('-2400000.5')

                    for row in flines[self._first_data:]:
                        photodict = OrderedDict()
                        rname = (row[cidict[ENTRY.NAME]]
                                 if ENTRY.NAME in cidict else event_names[0])
                        for pi in range(perms):
                            sources = set()
                            for key in cidict:
                                if key in self._bool_keys:
                                    rval = row[cidict[key]]

                                    if rval in self._FALSE_VALS:
                                        rval = False
                                    elif rval in self._TRUE_VALS:
                                        rval = True

                                    if type(rval) != 'bool':
                                        try:
                                            rval = bool(rval)
                                        except Exception:
                                            pass

                                    if type(rval) != 'bool':
                                        try:
                                            rval = bool(float(rval))
                                        except Exception:
                                            rval = True

                                    if not rval:
                                        continue
                                    if rval:
                                        row[cidict[key]] = rval
                                elif key == 'reference':
                                    if (isinstance(row[cidict[key]],
                                                   string_types)):
                                        srcdict = OrderedDict()
                                        if is_bibcode(row[cidict[key]]):
                                            srcdict[SOURCE.BIBCODE] = row[
                                                cidict[key]]
                                        else:
                                            srcdict[SOURCE.NAME] = row[
                                                cidict[key]]
                                        new_src = entries[rname].add_source(
                                            **srcdict)
                                        sources.update(new_src)
                                        # row[cidict[key]] = new_src
                                elif key == ENTRY.NAME:
                                    continue
                                elif (isinstance(key, Key) and
                                        key.type == KEY_TYPES.TIME and
                                        isinstance(cidict[key], list) and not
                                        isinstance(cidict[key],
                                                   string_types)):
                                    tval = np.array(row)[np.array(cidict[key][
                                        1:], dtype=int)]
                                    if cidict[key][0] == 'j':
                                        date = '-'.join([x.zfill(2) for x in
                                                         tval])
                                        date = self._month_rep.sub(
                                            lambda x: self._MONTH_IDS[
                                                x.group()], date)
                                        photodict[key] = str(
                                            astrotime(date, format='isot').mjd)
                                    elif cidict[key][0].upper() == 'JD':
                                        photodict[key] = str(
                                            jd_to_mjd(Decimal(tval[-1])))
                                    elif cidict[key][0].upper(
                                    ) == 'KILOSECONDS':
                                        photodict[key] = str(Decimal(
                                            KS_DAYS) * Decimal(tval[-1]))
                                    elif cidict[key][0].upper(
                                    ) == 'YYYY-MM-DD':
                                        photodict[key] = str(
                                            astrotime(
                                                tval[-1].replace('/', '-'),
                                                format='isot').mjd)
                                    continue

                                val = cidict[key]
                                if (isinstance(val, list) and not
                                        isinstance(val, string_types)):
                                    val = val[pi]
                                    if isinstance(val, string_types):
                                        if val != '':
                                            photodict[key] = val
                                    else:
                                        photodict[key] = row[val]
                                else:
                                    if isinstance(val, string_types):
                                        if val != '':
                                            photodict[key] = val
                                    else:
                                        photodict[key] = row[val]
                            if self._data_type == 2:
                                if self._zp:
                                    photodict[PHOTOMETRY.ZERO_POINT] = self._zp
                                else:
                                    photodict[PHOTOMETRY.ZERO_POINT] = (
                                        row[cidict[PHOTOMETRY.ZERO_POINT][pi]]
                                        if isinstance(cidict[
                                            PHOTOMETRY.ZERO_POINT], list) else
                                        row[cidict[PHOTOMETRY.ZERO_POINT]])
                                zpp = photodict[PHOTOMETRY.ZERO_POINT]
                                cc = (
                                    row[cidict[PHOTOMETRY.COUNT_RATE][pi]] if
                                    isinstance(cidict[
                                        PHOTOMETRY.COUNT_RATE], list) else
                                    row[cidict[PHOTOMETRY.COUNT_RATE]])
                                ecc = (
                                    row[cidict[PHOTOMETRY.E_COUNT_RATE][pi]] if
                                    isinstance(cidict[
                                        PHOTOMETRY.E_COUNT_RATE], list) else
                                    row[cidict[PHOTOMETRY.E_COUNT_RATE]])
                                if '<' in cc:
                                    set_pd_mag_from_counts(
                                        photodict, ec=cc.strip('<'), zp=zpp)
                                else:
                                    set_pd_mag_from_counts(
                                        photodict, c=cc, ec=ecc, zp=zpp)
                            elif self._data_type == 3:
                                photodict[
                                    PHOTOMETRY.U_FLUX_DENSITY] = self._ufd
                                if PHOTOMETRY.U_FLUX_DENSITY in cidict:
                                    photodict[PHOTOMETRY.U_FLUX_DENSITY] = (
                                        row[cidict[
                                            PHOTOMETRY.U_FLUX_DENSITY][pi]]
                                        if isinstance(cidict[
                                            PHOTOMETRY.
                                            U_FLUX_DENSITY], list) else
                                        row[cidict[PHOTOMETRY.U_FLUX_DENSITY]])
                                if photodict[
                                        PHOTOMETRY.U_FLUX_DENSITY] == '':
                                    photodict[
                                        PHOTOMETRY.U_FLUX_DENSITY] = 'µJy'
                                fd = (
                                    row[cidict[PHOTOMETRY.FLUX_DENSITY][pi]] if
                                    isinstance(cidict[
                                        PHOTOMETRY.FLUX_DENSITY], list) else
                                    row[cidict[PHOTOMETRY.FLUX_DENSITY]])
                                efd = (
                                    row[cidict[
                                        PHOTOMETRY.E_FLUX_DENSITY][pi]] if
                                    isinstance(cidict[
                                        PHOTOMETRY.E_FLUX_DENSITY], list) else
                                    row[cidict[PHOTOMETRY.E_FLUX_DENSITY]])

                                mult = Decimal('1')
                                ufd = photodict[PHOTOMETRY.U_FLUX_DENSITY]
                                if ufd.lower() in [
                                        'mjy', 'millijy', 'millijansky']:
                                    mult = Decimal('1e3')
                                elif ufd.lower() in ['jy', 'jansky']:
                                    mult = Decimal('1e6')

                                if '<' in fd:
                                    set_pd_mag_from_flux_density(
                                        photodict, efd=str(
                                            Decimal(fd.strip('<')) * mult))
                                else:
                                    set_pd_mag_from_flux_density(
                                        photodict, fd=Decimal(fd) * mult,
                                        efd=Decimal(efd) * mult)
                            if not len(sources) and not len(shared_sources):
                                if use_self_source is None:
                                    sopts = [
                                        ('Bibcode', 'b'), ('DOI', 'd'),
                                        ('ArXiv ID', 'a'), ('Last name', 'l')]
                                    if self._require_source:
                                        sel_str = 'must_select_source'
                                    else:
                                        sel_str = 'select_source'
                                    text = prt.text(sel_str)
                                    skind = prt.prompt(
                                        text, kind='option',
                                        options=sopts, default='b',
                                        none_string=(
                                            None if self._require_source else
                                            ('None of the above, '
                                             'tag MOSFiT as source')))
                                    if skind == 'b':
                                        rsource = OrderedDict()
                                        bibcode = ''

                                        while len(bibcode) != 19:
                                            bibcode = prt.prompt(
                                                'bibcode',
                                                kind='string',
                                                allow_blank=False
                                            )
                                            bibcode = bibcode.strip()
                                            if not is_bibcode(bibcode):
                                                bibcode = ''
                                        rsource[
                                            SOURCE.BIBCODE] = bibcode
                                        use_self_source = False
                                    elif skind == 'd':
                                        rsource = OrderedDict()
                                        doi = prt.prompt(
                                            'doi', kind='string',
                                            allow_blank=False)
                                        rsource[SOURCE.DOI] = doi.strip()
                                        use_self_source = False
                                    elif skind == 'a':
                                        rsource = OrderedDict()
                                        arxiv = prt.prompt(
                                            'arxiv', kind='string',
                                            allow_blank=False)
                                        rsource[SOURCE.ARXIVID] = arxiv.strip()
                                        use_self_source = False
                                    elif skind == 'l':
                                        rsource = OrderedDict()
                                        last_name = prt.prompt(
                                            'last_name', kind='string',
                                            allow_blank=False
                                        )
                                        rsource[
                                            SOURCE.NAME] = (
                                                last_name.strip().title() +
                                                ' et al., in preparation')
                                        use_self_source = False
                                    elif skind == 'n':
                                        use_self_source = True

                                    shared_sources.append(rsource)

                            if len(sources) or len(shared_sources):
                                src_list = list(sources)
                                for src in shared_sources:
                                    src_list.append(entries[
                                        rname].add_source(**src))
                                photodict[PHOTOMETRY.SOURCE] = ','.join(
                                    src_list)

                            if any([x in photodict.get(
                                    PHOTOMETRY.MAGNITUDE, '')
                                    for x in ['<', '>', '≤', '≥']]):
                                photodict[PHOTOMETRY.UPPER_LIMIT] = True
                                photodict[
                                    PHOTOMETRY.MAGNITUDE] = photodict[
                                        PHOTOMETRY.MAGNITUDE].strip('<>≤≥')

                            if '<' in photodict.get(PHOTOMETRY.COUNT_RATE, ''):
                                photodict[PHOTOMETRY.UPPER_LIMIT] = True
                                photodict[
                                    PHOTOMETRY.COUNT_RATE] = photodict[
                                        PHOTOMETRY.COUNT_RATE].strip('<')
                                if PHOTOMETRY.E_COUNT_RATE in photodict:
                                    del(photodict[PHOTOMETRY.E_COUNT_RATE])

                            if '<' in photodict.get(
                                    PHOTOMETRY.FLUX_DENSITY, ''):
                                photodict[PHOTOMETRY.UPPER_LIMIT] = True
                                photodict[
                                    PHOTOMETRY.FLUX_DENSITY] = photodict[
                                        PHOTOMETRY.FLUX_DENSITY].strip('<')
                                if PHOTOMETRY.E_FLUX_DENSITY in photodict:
                                    del(photodict[PHOTOMETRY.E_FLUX_DENSITY])

                            # Apply offset time if set.
                            if (PHOTOMETRY.TIME in photodict and
                                    toffset != Decimal('0')):
                                photodict[PHOTOMETRY.TIME] = str(
                                    Decimal(photodict[PHOTOMETRY.TIME]) +
                                    toffset)

                            # Remove some attributes if not numeric.
                            for attr in self._purge_non_numeric_keys:
                                if not is_number(photodict.get(attr, 0)):
                                    del(photodict[attr])

                            # Skip entries for which key values are not
                            # expected type.
                            if not all([
                                is_number(photodict.get(x, 0))
                                for x in photodict.keys() if
                                (PHOTOMETRY.get_key_by_name(x).type ==
                                 KEY_TYPES.NUMERIC)]):
                                continue

                            # Remove some attributes if not positive.
                            for attr in self._positive_keys:
                                if float(photodict.get(attr, 1)) <= 0:
                                    del(photodict[attr])

                            # Skip placeholder values.
                            if float(photodict.get(
                                    PHOTOMETRY.MAGNITUDE, 0.0)) > 50.0:
                                continue

                            # Add system if specified by user.
                            if (self._system is not None and
                                    PHOTOMETRY.SYSTEM not in photodict):
                                photodict[PHOTOMETRY.SYSTEM] = self._system

                            # Remove keys not in the `PHOTOMETRY` class.
                            for key in list(photodict.keys()):
                                if key not in PHOTOMETRY.vals():
                                    del(photodict[key])

                            # Add the photometry.
                            entries[rname].add_photometry(
                                **photodict)

                            # Add other entry keys.
                            for key in self._entry_keys:
                                if key not in cidict:
                                    continue
                                qdict = OrderedDict()
                                qdict[QUANTITY.VALUE] = row[
                                    cidict[key]]
                                if len(sources) or len(shared_sources):
                                    qdict[QUANTITY.SOURCE] = ','.join(
                                        src_list)
                                entries[rname].add_quantity(key, **qdict)

                    merge_with_existing = None
                    for ei, entry in enumerate(entries):
                        entries[entry].sanitize()
                        if os.path.isfile(new_events[ei]):
                            if merge_with_existing is None:
                                merge_with_existing = prt.prompt(
                                    'merge_with_existing', default='y')
                            if merge_with_existing:
                                print(event_names[ei])
                                existing = Entry.init_from_file(
                                    catalog=None,
                                    name=event_names[ei],
                                    path=new_events[ei],
                                    clean=True,
                                    merge=False,
                                    pop_schema=False,
                                    ignore_keys=[ENTRY.MODELS],
                                    compare_to_existing=False)
                                Catalog().copy_entry_to_entry(
                                    existing, entries[entry])

                        oentry = entries[entry]._ordered(entries[entry])
                        entabbed_json_dump(
                            {entry: oentry}, open(new_events[ei], 'w'),
                            separators=(',', ':'))

                    self._converted.extend([
                        [event_names[x], new_events[x]]
                        for x in range(len(event_names))])

                new_event_list.extend(new_events)
                previous_file = event
            else:
                new_event_list.append(event)

        return new_event_list

    def assign_columns(self, cidict, flines):
        """Assign columns based on header."""
        used_cis = OrderedDict()
        akeys = list(self._critical_keys) + list(self._helpful_keys)
        dkeys = list(self._dep_keys)
        prt = self._printer
        mpatt = re.compile('mag', re.IGNORECASE)

        for fi, fl in enumerate(flines):
            if not any([is_datum(x) for x in fl]):
                # Try to associate column names with common header keys.
                conflict_keys = []
                conflict_cis = []
                for ci, col in enumerate(fl):
                    for key in self._header_keys:
                        if any([(x[0] if isinstance(x, tuple)
                                 else x) == col.lower()
                                for x in self._header_keys[key]]):
                            if key in cidict or ci in used_cis:
                                # There is a conflict, ask user.
                                conflict_keys.append(key)
                                conflict_cis.append(ci)
                            else:
                                ind = [
                                    (x[0] if isinstance(x, tuple) else x)
                                    for x in self._header_keys[key]].index(
                                        col.lower())
                                match = self._header_keys[key][ind]
                                cidict[key] = [match[-1], ci] if isinstance(
                                    match, tuple) else ci
                                used_cis[ci] = key
                            break

                for cki, ck in enumerate(conflict_keys):
                    if ck in cidict:
                        ci = cidict[ck]
                        del(cidict[ck])
                        del(used_cis[ci if isinstance(
                            ci, string_types) else ci[-1]])
            else:
                self._first_data = fi
                break

        # Look for columns that are band names if no mag/counts/flux dens
        # column was found.
        if (not any([x in cidict for x in [
            PHOTOMETRY.MAGNITUDE, PHOTOMETRY.COUNT_RATE,
                PHOTOMETRY.FLUX_DENSITY]])):
            # Delete `E_MAGNITUDE` and `BAND` if they exist (we'll need to find
            # for each column).
            key = PHOTOMETRY.MAGNITUDE
            ekey = PHOTOMETRY.E_MAGNITUDE
            ukey = PHOTOMETRY.UPPER_LIMIT
            bkey = PHOTOMETRY.BAND
            if ekey in cidict:
                ci = cidict[ekey]
                del(cidict[used_cis[ci]])
                del(used_cis[ci])
            if ukey in cidict:
                ci = cidict[ukey]
                del(cidict[used_cis[ci]])
                del(used_cis[ci])
            if bkey in cidict:
                ci = cidict[bkey]
                del(cidict[used_cis[ci]])
                del(used_cis[ci])
            for fi, fl in enumerate(flines):
                if not any([is_datum(x) for x in fl]):
                    # Try to associate column names with common header keys.
                    for ci, col in enumerate(fl):
                        if ci in used_cis:
                            continue
                        ccol = mpatt.sub('', col)
                        if ccol in self._band_names:
                            cidict.setdefault(key, []).append(ci)
                            used_cis[ci] = key
                            cidict.setdefault(bkey, []).append(ccol)
                        elif ccol in self._emagstrs:
                            cidict.setdefault(ekey, []).append(ci)
                            used_cis[ci] = ekey
                        elif ccol in self._lmagstrs:
                            cidict.setdefault(ukey, []).append(ci)
                            used_cis[ci] = ukey

        # See which keys we collected. If we are missing any critical keys, ask
        # the user which column they are.

        # First ask the user if this data is in magnitudes or in counts.
        self._data_type = 1
        if (PHOTOMETRY.MAGNITUDE in cidict and
                PHOTOMETRY.COUNT_RATE not in cidict and
                PHOTOMETRY.FLUX_DENSITY not in cidict):
            self._data_type = 1
        elif (PHOTOMETRY.MAGNITUDE not in cidict and
              PHOTOMETRY.COUNT_RATE in cidict and
              PHOTOMETRY.FLUX_DENSITY not in cidict):
            self._data_type = 2
        elif (PHOTOMETRY.MAGNITUDE not in cidict and
              PHOTOMETRY.COUNT_RATE not in cidict and
              PHOTOMETRY.FLUX_DENSITY in cidict):
            self._data_type = 3
        else:
            self._data_type = prt.prompt(
                'counts_mags_fds', kind='option',
                options=['Magnitudes', 'Counts (per second)',
                         'Flux Densities (Jansky)'],
                none_string='No Photometry', default='1')
        if self._data_type in [1, 3, 'n']:
            akeys.remove(PHOTOMETRY.COUNT_RATE)
            akeys.remove(PHOTOMETRY.E_COUNT_RATE)
            akeys.remove(PHOTOMETRY.ZERO_POINT)
            dkeys.remove(PHOTOMETRY.E_COUNT_RATE)
        if self._data_type in [2, 3, 'n']:
            akeys.remove(PHOTOMETRY.MAGNITUDE)
            akeys.remove(PHOTOMETRY.E_MAGNITUDE)
            dkeys.remove(PHOTOMETRY.E_MAGNITUDE)
        if self._data_type in [1, 2, 'n']:
            akeys.remove(PHOTOMETRY.FLUX_DENSITY)
            akeys.remove(PHOTOMETRY.E_FLUX_DENSITY)
            if (PHOTOMETRY.E_LOWER_FLUX_DENSITY in cidict and
                    PHOTOMETRY.E_UPPER_FLUX_DENSITY in cidict):
                akeys.remove(PHOTOMETRY.E_FLUX_DENSITY)
            dkeys.remove(PHOTOMETRY.E_FLUX_DENSITY)
            dkeys.remove(PHOTOMETRY.U_FLUX_DENSITY)
        if self._data_type == 'n':
            akeys.remove(PHOTOMETRY.TIME)
            akeys.remove(PHOTOMETRY.BAND)
            akeys.remove(PHOTOMETRY.INSTRUMENT)
            akeys.remove(PHOTOMETRY.TELESCOPE)
            akeys.append(ENTRY.NAME)

        # Make sure `E_` keys always appear after the actual measurements.
        if (PHOTOMETRY.MAGNITUDE in akeys and
                PHOTOMETRY.E_MAGNITUDE in akeys):
            akeys.remove(PHOTOMETRY.E_MAGNITUDE)
            akeys.insert(
                akeys.index(PHOTOMETRY.MAGNITUDE) + 1,
                PHOTOMETRY.E_MAGNITUDE)
        # Remove regular `E_` keys if both `E_LOWER_`/`E_UPPER_` exist.
        if (PHOTOMETRY.E_LOWER_MAGNITUDE in cidict and
                PHOTOMETRY.E_UPPER_MAGNITUDE in cidict):
            akeys.remove(PHOTOMETRY.E_MAGNITUDE)

        columns = np.array(flines[self._first_data:]).T.tolist()
        colstrs = np.array([
            ', '.join(x[:5]) + ', ...' for x in columns])
        colinds = np.setdiff1d(np.arange(
            len(colstrs)), list([x[-1] if (
                isinstance(x, list) and not isinstance(
                    x, string_types)) else x for x in cidict.values()]))
        ignore = prt.message('ignore_column', prt=False)
        specify = prt.message('specify_column', prt=False)
        for key in akeys:
            selected_cols = [
                y for y in [a for b in [
                    listify(x) for x in list(cidict.values())] for a in b]
                if isinstance(y, (int, np.integer))]
            if key in cidict:
                continue
            if key in dkeys and self._use_mc:
                continue
            if key.type == KEY_TYPES.NUMERIC:
                lcolinds = [x for x in colinds
                            if any(is_datum(y) for y in columns[x]) and
                            x not in selected_cols]
            elif key.type == KEY_TYPES.TIME:
                lcolinds = [x for x in colinds
                            if any(is_date(y) or is_datum(y)
                                   for y in columns[x]) and
                            x not in selected_cols]
            elif key.type == KEY_TYPES.STRING:
                lcolinds = [x for x in colinds
                            if any(not is_datum(y) for y in columns[x]) and
                            x not in selected_cols]
            else:
                lcolinds = [x for x in colinds if x not in selected_cols]
            select = False
            selects = []
            while select is False:
                mc = 1
                if key in self._mc_keys:
                    pkey = self._inflect.plural(key)
                    text = prt.message(
                        'one_per_line', [key, pkey, pkey],
                        prt=False)
                    mc = prt.prompt(
                        text, kind='option', message=False,
                        none_string=None,
                        options=[
                            'One `{}` per row'.format(key),
                            'Multiple `{}` per row'.format(pkey)])
                if mc == 1:
                    text = prt.message(
                        'no_matching_column', [key], prt=False)
                    ns = (
                        ignore if key in (
                            self._optional_keys + self._helpful_keys) else
                        specify if key in self._specify_keys
                        else None)
                    if len(colstrs[lcolinds]):
                        select = prt.prompt(
                            text, message=False,
                            kind='option', none_string=ns,
                            default=('j' if ns is None and
                                     len(colstrs[lcolinds]) > 1
                                     else None if ns is None else 'n'),
                            options=colstrs[lcolinds].tolist() + (
                                [('Multiple columns need to be joined.', 'j')]
                                if len(colstrs[lcolinds]) > 1 else []))
                    else:
                        select = None
                    if select == 'j':
                        select = None
                        jsel = None
                        selects.append('j')
                        while jsel != 'd' and len(lcolinds):
                            jsel = prt.prompt(
                                'join_which_columns', default='d',
                                kind='option', none_string=None,
                                options=colstrs[lcolinds].tolist() + [
                                    ('All columns to be joined '
                                     'have been selected.', 'd')
                                ])
                            if jsel != 'd':
                                selects.append(lcolinds[jsel - 1])
                                lcolinds = np.delete(lcolinds, jsel - 1)
                else:
                    self._use_mc = True
                    select = False
                    while select is not None and len(lcolinds):
                        text = prt.message(
                            'select_mc_column', [key], prt=False)
                        select = prt.prompt(
                            text, message=False,
                            kind='option', default='n',
                            none_string='No more `{}` columns.'.format(key),
                            options=colstrs[lcolinds].tolist())
                        if select is not None and select is not False:
                            selects.append(lcolinds[select - 1])
                            lcolinds = np.delete(lcolinds, select - 1)
                        else:
                            break
                        for dk in dkeys:
                            dksel = None
                            while dksel is None:
                                text = prt.message(
                                    'select_dep_column', [dk, key], prt=False)
                                sk = dk in self._specify_keys
                                if not sk:
                                    dksel = prt.prompt(
                                        text, message=False,
                                        kind='option', none_string=None,
                                        options=colstrs[lcolinds].tolist())
                                    if dksel is not None:
                                        selects.append(lcolinds[dksel - 1])
                                        lcolinds = np.delete(
                                            lcolinds, dksel - 1)
                                else:
                                    spectext = prt.message(
                                        'specify_mc_value', [dk, key],
                                        prt=False)
                                    val = ''
                                    while val.strip() is '':
                                        val = prt.prompt(
                                            spectext, message=False,
                                            kind='string')
                                    selects.append(val)
                                    break

            if select is not None:
                iselect = int(select)
                cidict[key] = lcolinds[iselect - 1]
                colinds = np.delete(colinds, np.argwhere(
                    colinds == lcolinds[iselect - 1]))
            elif len(selects):
                if selects[0] == 'j':
                    cidict[key] = selects
                else:
                    kdkeys = [key] + dkeys
                    allk = list(OrderedDict.fromkeys(kdkeys).keys())
                    for ki, k in enumerate(allk):
                        cidict[k] = [
                            colinds[s - 1] if isinstance(s, (
                                int, np.integer)) else s
                            for s in selects[ki::len(allk)]]
                    for s in selects:
                        if not isinstance(s, (int, np.integer)):
                            continue
                        colinds = np.delete(colinds, np.argwhere(
                            colinds == s - 1))
            elif key in self._specify_keys:
                msg = ('specify_value_blank' if key in self._helpful_keys else
                       'specify_value')
                text = prt.message(msg, [key], prt=False)
                cidict[key] = prt.prompt(
                    text, message=False, kind='string', allow_blank=(
                        key in self._helpful_keys))

        self._zp = ''
        if self._data_type == 2 and PHOTOMETRY.ZERO_POINT not in cidict:
            while not is_number(self._zp):
                self._zp = prt.prompt('zeropoint', kind='string')

        self._ufd = None
        if self._data_type == 3 and PHOTOMETRY.U_FLUX_DENSITY not in cidict:
            while ((self._ufd.lower() if self._ufd is not None else None)
                   not in ['µjy', 'mjy', 'jy', 'microjy', 'millijy', 'jy',
                           'microjansky', 'millijansky', 'jansky', '']):
                self._ufd = prt.prompt('u_flux_density', kind='string')

        self._system = None
        if self._data_type == 1 and PHOTOMETRY.SYSTEM not in cidict:
            systems = ['AB', 'Vega']
            self._system = prt.prompt(
                'system', kind='option', options=systems,
                none_string='Use default for all bands.',
                default='n')
            if self._system is not None:
                self._system = systems[int(self._system) - 1]

        if (PHOTOMETRY.INSTRUMENT not in cidict and
            PHOTOMETRY.TELESCOPE not in cidict and
                PHOTOMETRY.MAGNITUDE in cidict):
            prt.message('instrument_recommended', warning=True)

    def get_converted(self):
        """Get a list of events that were converted from ASCII to JSON."""
        return self._converted



"""Definitions for the `NickelCobalt` class."""
import numpy as np

from mosfit.modules.engines.engine import Engine


# Important: Only define one ``Module`` class per file.


class NickelCobalt(Engine):
    """Nickel/Cobalt decay engine."""

    NI56_LUM = 6.45e43
    CO56_LUM = 1.45e43
    NI56_LIFE = 8.8
    CO56_LIFE = 111.3

    def process(self, **kwargs):
        """Process module."""
        self._times = kwargs[self.key('dense_times')]
        self._mnickel = kwargs[self.key('fnickel')] * kwargs[
            self.key('mejecta')]
        self._rest_t_explosion = kwargs[self.key('resttexplosion')]

        # From 1994ApJS...92..527N
        ts = np.empty_like(self._times)
        t_inds = self._times >= self._rest_t_explosion
        ts[t_inds] = self._times[t_inds] - self._rest_t_explosion

        luminosities = np.zeros_like(self._times)
        luminosities[t_inds] = self._mnickel * (
            self.NI56_LUM * np.exp(-ts[t_inds] / self.NI56_LIFE) +
            self.CO56_LUM * np.exp(-ts[t_inds] / self.CO56_LIFE))

        luminosities[np.isnan(luminosities)] = 0.0

        return {self.dense_key('luminosities'): luminosities}



"""Definitions for the `MultiBlackbody` class."""
from math import pi

import numexpr as ne
import numpy as np
from astropy import constants as c
from mosfit.constants import DAY_CGS, FOUR_PI, KM_CGS, M_SUN_CGS  # noqa: F401
from mosfit.modules.seds.sed import SED


# Important: Only define one ``Module`` class per file.


class MultiBlackbody(SED):
    """Generalized multiple blackbody spectral energy distribution."""

    FLUX_CONST = FOUR_PI * (2.0 * c.h / (c.c ** 2) * pi).cgs.value
    X_CONST = (c.h / c.k_B).cgs.value
    STEF_CONST = (4.0 * pi * c.sigma_sb).cgs.value

    def process(self, **kwargs):
        """Process module."""
        raise RuntimeError('`MultiBlackbody` is not yet functional.')
        kwargs = self.prepare_input(self.key('luminosities'), **kwargs)
        self._luminosities = kwargs[self.key('luminosities')]
        self._bands = kwargs['all_bands']
        self._band_indices = kwargs['all_band_indices']
        self._areas = kwargs[self.key('areas')]
        self._temperature_phots = kwargs[self.key('temperaturephots')]
        xc = self.X_CONST  # noqa: F841
        fc = self.FLUX_CONST  # noqa: F841
        temperature_phot = self._temperature_phot
        zp1 = 1.0 + kwargs[self.key('redshift')]
        seds = []
        for li, lum in enumerate(self._luminosities):
            cur_band = self._bands[li]  # noqa: F841
            bi = self._band_indices[li]
            rest_freqs = [x * zp1  # noqa: F841
                          for x in self._sample_frequencies[bi]]
            wav_arr = np.array(self._sample_wavelengths[bi])  # noqa: F841
            radius_phot = self._radius_phot[li]  # noqa: F841
            temperature_phot = self._temperature_phot[li]  # noqa: F841

            if li == 0:
                sed = ne.evaluate(
                    'fc * radius_phot**2 * rest_freqs**3 / '
                    '(exp(xc * rest_freqs / temperature_phot) - 1.0)')
            else:
                sed = ne.re_evaluate()

            sed = np.nan_to_num(sed)

            seds.append(list(sed))

        seds = self.add_to_existing_seds(seds, **kwargs)

        # Units of `seds` is ergs / s / Angstrom.
        return {'sample_wavelengths': self._sample_wavelengths,
                self.key('seds'): seds}



"""Definitions for the `RProcess` class."""
from math import isnan

import numpy as np
from astrocats.catalog.source import SOURCE
from mosfit.constants import C_CGS, DAY_CGS, IPI, KM_CGS, M_SUN_CGS
from mosfit.modules.engines.engine import Engine
from scipy.interpolate import RegularGridInterpolator


# Important: Only define one ``Module`` class per file.


class RProcess(Engine):
    """r-process decay engine.

    input luminosity adapted from Metzger 2016: 2017LRR....20....3M
    """

    _REFERENCES = [
        {SOURCE.BIBCODE: '2013ApJ...775...18B'},
        {SOURCE.BIBCODE: '2017LRR....20....3M'},
        {SOURCE.BIBCODE: '2017arXiv170708132V'}
    ]

    ckm = C_CGS / KM_CGS

    def __init__(self, **kwargs):
        """Initialize module."""
        super(Engine, self).__init__(**kwargs)
        self._wants_dense = True
        barnes_v = np.asarray([0.1, 0.2, 0.3])
        barnes_M = np.asarray([1.e-3, 5.e-3, 1.e-2, 5.e-2])
        barnes_a = np.asarray([[2.01, 4.52, 8.16], [0.81, 1.9, 3.2], [
                              0.56, 1.31, 2.19], [.27, .55, .95]])
        barnes_b = np.asarray([[0.28, 0.62, 1.19], [0.19, 0.28, 0.45], [
                              0.17, 0.21, 0.31], [0.10, 0.13, 0.15]])
        barnes_d = np.asarray([[1.12, 1.39, 1.52], [0.86, 1.21, 1.39], [
                              0.74, 1.13, 1.32], [0.6, 0.9, 1.13]])

        self.therm_func_a = RegularGridInterpolator(
            (barnes_M, barnes_v), barnes_a, bounds_error=False, fill_value=None)
        self.therm_func_b = RegularGridInterpolator(
            (barnes_M, barnes_v), barnes_b, bounds_error=False, fill_value=None)
        self.therm_func_d = RegularGridInterpolator(
            (barnes_M, barnes_v), barnes_d, bounds_error=False, fill_value=None)

    def process(self, **kwargs):
        """Process module."""
        self._times = kwargs[self.key('dense_times')]
        self._mass = kwargs[self.key('mejecta')] * M_SUN_CGS
        self._rest_texplosion = kwargs[self.key('resttexplosion')]
        self._vejecta = kwargs[self.key('vejecta')]
        self._a = self.therm_func_a(
            [self._mass / M_SUN_CGS, self._vejecta / self.ckm])[0]
        self._bx2 = 2.0 * self.therm_func_b(
            [self._mass / M_SUN_CGS, self._vejecta / self.ckm])[0]
        self._d = self.therm_func_d(
            [self._mass / M_SUN_CGS, self._vejecta / self.ckm])[0]

        ts = [
            np.inf
            if self._rest_texplosion > x else (x - self._rest_texplosion)
            for x in self._times
        ]

        self._lscale = self._mass * 4.0e18 * 0.36
        luminosities = [
            self._lscale * (0.5 - IPI * np.arctan(
                (t * DAY_CGS - 1.3) / 0.11)) ** 1.3 *
            (np.exp(-self._a * t) + np.log1p(
                self._bx2 * t ** self._d) / (self._bx2 * t ** self._d))
            for t in ts
        ]
        luminosities = [0.0 if isnan(x) else x for x in luminosities]

        return {self.dense_key('luminosities'): luminosities}


"""Definitions for the `Fallback` class."""
import os

import astropy.constants as c
import numpy as np
from scipy.interpolate import interp1d

from mosfit.constants import C_CGS, DAY_CGS, FOUR_PI, M_SUN_CGS
from mosfit.modules.engines.engine import Engine

CLASS_NAME = 'Fallback'


class Fallback(Engine):
    """A tde engine."""

    def __init__(self, **kwargs):
        """Initialize module.

        Loads and interpolates tde simulation data. Simulation data is
        from Guillochon 2013 and can be found on astrocrash.net.
        The files in data directory have been converted from dm/dt space
        to dm/de space.
        """
        super(Fallback, self).__init__(**kwargs)

        G = c.G.cgs.value  # 6.67259e-8 cm3 g-1 s-2
        Mhbase = 1.0e6 * M_SUN_CGS  # this is the generic size of bh used

        self.EXTRAPOLATE = True

        # ------ DIRECTORY PARAMETERS -------

        # It is assumed that there are different files for each beta
        # (such as 2.500.dat for beta = 2.5)
        # The first row is energy, the second is dmde.

        self._gammas = ['4-3', '5-3']

        # dictionaries with gamma's as keys.
        self._beta_slope = {self._gammas[0]: [], self._gammas[1]: []}
        self._beta_yinter = {self._gammas[0]: [], self._gammas[1]: []}
        self._sim_beta = {self._gammas[0]: [], self._gammas[1]: []}
        self._mapped_time = {self._gammas[0]: [], self._gammas[1]: []}
        # for converting back from mapped time to actual times and doing
        # interpolation in actual time
        self._premaptime = {self._gammas[0]: [], self._gammas[1]: []}
        self._premapdmdt = {self._gammas[0]: [], self._gammas[1]: []}

        for g in self._gammas:

            dmdedir = (os.path.dirname(__file__)[:-15] + 'models/tde/data/' +
                       g + '/')

            # --------- GET SIMULATION BETAS -----------------
            sim_beta_files = os.listdir(dmdedir)

            self._sim_beta[g].extend([float(b[:-4]) for b in sim_beta_files])

            # ----- CREATE INTERPOLATION FUNCTIONS; FIND SLOPES & YINTERs -----
            time = {}
            dmdt = {}
            ipeak = {}
            mapped_time = {}
            # get dmdt and t for the lowest beta value
            # energy & dmde (cgs)
            e, d = np.loadtxt(dmdedir + sim_beta_files[0])
            # only convert dm/de --> dm/dt for mass that is bound to BH (e < 0)
            ebound = e[e < 0]
            dmdebound = d[e < 0]

            if min(dmdebound) < 0:  # shouldn't happen, just a check

                print('beta, gamma, negative dmde bound:', self._sim_beta[g],
                      g, dmdebound[dmdebound < 0])

            # calculate de/dt, time and dm/dt arrays
            # de/dt in log(/s), time in log(seconds), dm/dt in log(g/s)
            dedt = (1.0 / 3.0) * (-2.0 * ebound) ** (5.0 / 2.0) / \
                (2.0 * np.pi * G * Mhbase)
            time['lo'] = np.log10((2.0 * np.pi * G * Mhbase) *
                                  (-2.0 * ebound) ** (-3.0 / 2.0))
            dmdt['lo'] = np.log10(dmdebound * dedt)

            ipeak['lo'] = np.argmax(dmdt['lo'])

            # split time['lo'] & dmdt['lo'] into pre-peak and post-peak array
            time['lo'] = np.array([
                time['lo'][:ipeak['lo']],
                time['lo'][ipeak['lo']:]])  # peak in array 2
            dmdt['lo'] = np.array([
                dmdt['lo'][:ipeak['lo']],
                dmdt['lo'][ipeak['lo']:]])  # peak in array 2

            # will contain time/dmdt arrays
            # (split into pre & post peak times/dmdts)
            # for each beta value
            self._premaptime[g].append(np.copy(time['lo']))
            self._premapdmdt[g].append(np.copy(dmdt['lo']))

            for i in range(1, len(self._sim_beta[g])):
                # indexing this way bc calculating slope and yintercepts
                # BETWEEN each simulation beta

                e, d = np.loadtxt(dmdedir + sim_beta_files[i])
                # only convert dm/de --> dm/dt for mass bound to BH (e < 0)
                ebound = e[e < 0]
                dmdebound = d[e < 0]

                if min(dmdebound) < 0:  # shouldn't happen, just a check
                    print('beta, gamma, negative dmde bound:',
                          self._sim_beta[g], g, dmdebound[dmdebound < 0])

                # calculate de/dt, time and dm/dt arrays
                # de/dt in log(erg/s), time in log(seconds), dm/dt in log(g/s)
                dedt = (1.0 / 3.0) * (-2.0 * ebound) ** (5.0 / 2.0) / \
                    (2.0 * np.pi * G * Mhbase)
                time['hi'] = np.log10((2.0 * np.pi * G * Mhbase) *
                                      (-2.0 * ebound) ** (-3.0 / 2.0))
                dmdt['hi'] = np.log10(dmdebound * dedt)

                ipeak['hi'] = np.argmax(dmdt['hi'])

                # split time_hi and dmdt_hi into pre-peak and post-peak array
                # peak in 2nd array
                time['hi'] = np.array([time['hi'][:ipeak['hi']],
                                       time['hi'][ipeak['hi']:]])
                dmdt['hi'] = np.array([dmdt['hi'][:ipeak['hi']],
                                       dmdt['hi'][ipeak['hi']:]])
                # will contain time/dmdt arrays
                # (split into pre & post peak times/dmdts)
                # for each beta value
                self._premapdmdt[g].append(np.copy(dmdt['hi']))
                self._premaptime[g].append(np.copy(time['hi']))

                mapped_time['hi'] = []
                mapped_time['lo'] = []

                self._beta_slope[g].append([])
                self._beta_yinter[g].append([])
                self._mapped_time[g].append([])
                for j in [0, 1]:  # once before peak, once after peak
                    # choose more densely sampled curve to map times to 0-1
                    # less densely sampled curve will be interpolated to match
                    if len(time['lo'][j]) < len(time['hi'][j]):
                        # hi array more densely sampled
                        interp = 'lo'
                        nointerp = 'hi'
                    else:
                        # will also catch case where they have the same lengths
                        interp = 'hi'
                        nointerp = 'lo'
                    # map times from more densely sampled curves
                    # (both pre & post peak, might be from diff. dmdts)
                    # to 0 - 1
                    mapped_time[nointerp].append(
                        1. / (time[nointerp][j][-1] - time[nointerp][j][0]) *
                        (time[nointerp][j] - time[nointerp][j][0]))
                    mapped_time[interp].append(
                        1. / (time[interp][j][-1] - time[interp][j][0]) *
                        (time[interp][j] - time[interp][j][0]))

                    # ensure bounds are same for interp and nointerp
                    # before interpolation
                    # (should be 0 and 1 from above, but could be slightly off
                    # due to rounding errors in python)
                    mapped_time[interp][j][0] = 0
                    mapped_time[interp][j][-1] = 1
                    mapped_time[nointerp][j][0] = 0
                    mapped_time[nointerp][j][-1] = 1

                    func = interp1d(mapped_time[interp][j], dmdt[interp][j])
                    dmdtinterp = func(mapped_time[nointerp][j])

                    if interp == 'hi':
                        slope = ((dmdtinterp - dmdt['lo'][j]) /
                                 (self._sim_beta[g][i] - self._sim_beta[g][
                                     i - 1]))
                    else:
                        slope = ((dmdt['hi'][j] - dmdtinterp) /
                                 (self._sim_beta[g][i] - self._sim_beta[g][
                                     i - 1]))
                    self._beta_slope[g][-1].append(slope)

                    yinter1 = (dmdt[nointerp][j] - self._beta_slope[g][-1][j] *
                               self._sim_beta[g][i - 1])
                    yinter2 = (dmdtinterp - self._beta_slope[g][-1][j] *
                               self._sim_beta[g][i])
                    self._beta_yinter[g][-1].append((yinter1 + yinter2) / 2.0)
                    self._mapped_time[g][-1].append(
                        np.array(mapped_time[nointerp][j]))

                time['lo'] = np.copy(time['hi'])
                dmdt['lo'] = np.copy(dmdt['hi'])

    def process(self, **kwargs):
        """Process module."""
        beta_interp = True
        beta_outside_range = False

        Mhbase = 1.0e6  # in units of Msolar, this is generic Mh used
        # in astrocrash sims
        Mstarbase = 1.0  # in units of Msolar
        Rstarbase = 1.0  # in units of Rsolar

        # this is not beta, but rather a way to map beta_4-3 --> beta_5-3
        # b = 0 --> min disruption, b = 1 --> full disruption,
        # b = 2 --> max beta of sims
        self._b = kwargs['b']

        if 0 <= self._b < 1:
            # 0.6 is min disruption beta for gamma = 4/3
            # 1.85 is full disruption beta for gamma = 4/3
            beta43 = 0.6 + 1.25 * self._b  # 0.6 + (1.85 - 0.6)*b
            # 0.5 is min disruption beta for gamma = 5/3
            # 0.9 is full disruption beta for gamma = 5/3
            beta53 = 0.5 + 0.4 * self._b  # 0.5 + (0.9 - 0.5)*b

            self._betas = {'4-3': beta43, '5-3': beta53}

        elif 1 <= self._b <= 2:
            beta43 = 1.85 + 2.15 * (self._b - 1)
            beta53 = 0.9 + 1.6 * (self._b - 1)
            self._betas = {'4-3': beta43, '5-3': beta53}

        else:
            self._printer.prt(
                'b outside range, bmin = 0; bmax = 2; b = {}'.format(
                    self._b))
            self._b = 2.0 if self._b > 2 else 0.0
            beta_outside_range = True

        # GET GAMMA VALUE

        gamma_interp = False

        self._Mstar = kwargs.get(self.key('starmass'))
        if self._Mstar <= 0.3 or self._Mstar >= 22:
            gammas = [self._gammas[1]]  # gamma = ['5-3']
            self._beta = self._betas['5-3']
        elif 1 <= self._Mstar <= 15:
            gammas = [self._gammas[0]]  # gamma = ['4-3']
            self._beta = self._betas['4-3']
        elif 0.3 < self._Mstar < 1:
            # region going from gamma = 5/3 to gamma = 4/3 as mass increases
            gamma_interp = True
            gammas = self._gammas
            # gfrac should == 0 for 4/3; == 1 for 5/3
            gfrac = (self._Mstar - 1.) / (0.3 - 1.)
            # beta_43 is always larger than beta_53
            self._beta = self._betas['5-3'] + (
                self._betas['4-3'] - self._betas['5-3']) * (1. - gfrac)
        elif 15 < self._Mstar < 22:
            # region going from gamma = 4/3 to gamma = 5/3 as mass increases
            gamma_interp = True
            gammas = self._gammas
            # gfrac should == 0 for 4/3; == 1 for 5/3
            gfrac = (self._Mstar - 15.) / (22. - 15.)

            # beta_43 is always larger than beta_53
            self._beta = self._betas['5-3'] + (
                self._betas['4-3'] - self._betas['5-3']) * (1. - gfrac)

        # try decoupling gamma from starmass
        '''
        self._scaled_gamma = kwargs['scaledgamma']
        # print (self._scaled_gamma)
        if self._scaled_gamma == 0.0: gammas = [self._gammas[0]]
        elif self._scaled_gamma == 1.0: gammas = [self._gammas[1]]
        else:
            gamma_interp = True
            gammas = self._gammas
            gfrac = self._scaled_gamma
        '''
        timedict = {}  # will hold time arrays for each g in gammas
        dmdtdict = {}  # will hold dmdt arrays for each g in gammas

        for g in gammas:
            # find simulation betas to interpolate between
            for i in range(len(self._sim_beta[g])):
                if self._betas[g] == self._sim_beta[g][i]:
                    # no need to interp, already have dmdt & t for this beta
                    beta_interp = False
                    interp_index_low = i
                    break

                if self._betas[g] < self._sim_beta[g][i]:
                    interp_index_high = i
                    interp_index_low = i - 1
                    beta_interp = True
                    break

            if beta_interp:
                # ----------- LINEAR BETA INTERPOLATION --------------

                # get new dmdts  (2 arrays, pre & post peak (peak in array 2))
                # use interp_index_low bc of how slope and yintercept are saved
                # (slope[0] corresponds to between beta[0] and beta[1] etc.)
                dmdt = np.array([
                    self._beta_yinter[g][interp_index_low][0] +
                    self._beta_slope[g][interp_index_low][0] * self._betas[g],
                    self._beta_yinter[g][interp_index_low][1] +
                    self._beta_slope[g][interp_index_low][1] * self._betas[g]])

                # map mapped_times back to actual times, requires interpolation
                # in time
                # first for pre peak times

                time = []
                for i in [0, 1]:
                    # interp_index_low indexes beta
                    # mapped time between beta low and beta high
                    time_betalo = (
                        self._mapped_time[g][interp_index_low][i] *
                        (self._premaptime[g][interp_index_low][i][-1] -
                         self._premaptime[g][interp_index_low][i][0]) +
                        self._premaptime[g][interp_index_low][i][0])
                    time_betahi = (
                        self._mapped_time[g][interp_index_low][i] *
                        (self._premaptime[g][interp_index_high][i][-1] -
                         self._premaptime[g][interp_index_high][i][0]) +
                        self._premaptime[g][interp_index_high][i][0])

                    time.append(
                        time_betalo + (time_betahi - time_betalo) *
                        (self._betas[g] -
                         self._sim_beta[g][interp_index_low]) /
                        (self._sim_beta[g][interp_index_high] -
                         self._sim_beta[g][interp_index_low]))

                time = np.array(time)

                timedict[g] = time
                dmdtdict[g] = dmdt

            elif not beta_interp:
                timedict[g] = np.copy(self._premaptime[g][interp_index_low])
                dmdtdict[g] = np.copy(self._premapdmdt[g][interp_index_low])

        # ---------------- GAMMA INTERPOLATION -------------------

        if gamma_interp:

            mapped_time = {'4-3': [], '5-3': []}

            time = []
            dmdt = []
            for j in [0, 1]:  # once before peak, once after peak
                # choose more densely sampled curve to map times to 0-1
                # less densely sampled curve will be interpolated to match
                if len(timedict['4-3'][j]) < len(timedict['5-3'][j]):
                    # gamma = 5/3 array more densely sampled
                    interp = '4-3'
                    nointerp = '5-3'
                else:
                    # will also catch case where they have the same lengths
                    interp = '5-3'
                    nointerp = '4-3'

                # map times from more densely sampled curves
                # (both pre & post peak, might be from diff. dmdts)
                # to 0 - 1
                mapped_time[nointerp].append(
                    1. / (timedict[nointerp][j][-1] -
                          timedict[nointerp][j][0]) *
                    (timedict[nointerp][j] - timedict[nointerp][j][0]))
                mapped_time[interp].append(
                    1. / (timedict[interp][j][-1] - timedict[interp][j][0]) *
                    (timedict[interp][j] - timedict[interp][j][0]))
                # ensure bounds same for interp & nointerp before interpolation
                # (they should be 0 and 1 from above, but could be slightly off
                # due to rounding errors in python)
                mapped_time[interp][j][0] = 0
                mapped_time[interp][j][-1] = 1
                mapped_time[nointerp][j][0] = 0
                mapped_time[nointerp][j][-1] = 1

                func = interp1d(mapped_time[interp][j], dmdtdict[interp][j])
                dmdtdict[interp][j] = func(mapped_time[nointerp][j])

                # recall gfrac = 0 --> gamma = 4/3, gfrac = 1 --> gamma 5/3
                if interp == '5-3':
                    # then mapped_time = mapped_time[nointerp] =
                    # mapped_time['4-3']
                    time53 = (mapped_time['4-3'][j] * (timedict['5-3'][j][-1] -
                                                       timedict['5-3'][j][0]) +
                              timedict['5-3'][j][0])
                    # convert back from logspace before adding to time array
                    time.extend(10 ** (timedict['4-3'][j] +
                                       (time53 - timedict['4-3'][j]) * gfrac))
                else:
                    # interp == '4-3'
                    time43 = (mapped_time['5-3'][j] * (timedict['4-3'][j][-1] -
                                                       timedict['4-3'][j][0]) +
                              timedict['4-3'][j][0])
                    # convert back from logspace before adding to time array
                    time.extend(10 ** (time43 +
                                       (timedict['5-3'][j] - time43) * gfrac))

                # recall gfrac = 0 --> gamma = 4/3, gfrac = 1 --> gamma 5/3
                # convert back from logspace before adding to dmdt array
                dmdt.extend(10 ** (dmdtdict['4-3'][j] +
                                   (dmdtdict['5-3'][j] -
                                    dmdtdict['4-3'][j]) * gfrac))

        else:  # gamma_interp == False
            # in this case, g will still be g from loop over gammas,
            # but there was only one gamma (no interpolation),
            # so g is the correct gamma
            # note that timedict[g] is a list not an array
            # no longer need a prepeak and postpeak array
            time = np.concatenate((timedict[g][0], timedict[g][1]))
            time = 10 ** time
            dmdt = np.concatenate((dmdtdict[g][0], dmdtdict[g][1]))
            dmdt = 10 ** dmdt

        time = np.array(time)
        dmdt = np.array(dmdt)

        # ----------- SCALE dm/dt TO BH & STAR MASS & STAR RADIUS -------------

        if 'dense_times' in kwargs:
            self._times = kwargs['dense_times']  # time in days
        else:
            print('in fallback, dense_times NOT in kwargs')
            self._times = kwargs['rest_times']

        # bh mass for dmdt's in astrocrash is 1e6 solar masses
        # dmdt ~ Mh^(-1/2)
        self._Mh = kwargs['bhmass']  # in units of solar masses

        # Assume that BDs below 0.1 solar masses are n=1 polytropes
        if self._Mstar < 0.1:
            Mstar_Tout = 0.1
        else:
            Mstar_Tout = self._Mstar

        # calculate Rstar from Mstar (using Tout et. al. 1996),
        # in Tout paper -> Z = 0.02 (now not quite solar Z) and ZAMS
        Z = 0.0134  # assume solar metallicity
        log10_Z_02 = np.log10(Z / 0.02)

        # Tout coefficients for calculating Rstar
        Tout_theta = (1.71535900 + 0.62246212 * log10_Z_02 - 0.92557761 *
                      log10_Z_02 ** 2 - 1.16996966 * log10_Z_02 ** 3 -
                      0.30631491 *
                      log10_Z_02 ** 4)
        Tout_l = (6.59778800 - 0.42450044 * log10_Z_02 - 12.13339427 *
                  log10_Z_02 ** 2 - 10.73509484 * log10_Z_02 ** 3 -
                  2.51487077 * log10_Z_02 ** 4)
        Tout_kpa = (10.08855000 - 7.11727086 * log10_Z_02 - 31.67119479 *
                    log10_Z_02 ** 2 - 24.24848322 * log10_Z_02 ** 3 -
                    5.33608972 * log10_Z_02 ** 4)
        Tout_lbda = (1.01249500 + 0.32699690 * log10_Z_02 - 0.00923418 *
                     log10_Z_02 ** 2 - 0.03876858 * log10_Z_02 ** 3 -
                     0.00412750 * log10_Z_02 ** 4)
        Tout_mu = (0.07490166 + 0.02410413 * log10_Z_02 + 0.07233664 *
                   log10_Z_02 ** 2 + 0.03040467 * log10_Z_02 ** 3 +
                   0.00197741 * log10_Z_02 ** 4)
        Tout_nu = 0.01077422
        Tout_eps = (3.08223400 + 0.94472050 * log10_Z_02 - 2.15200882 *
                    log10_Z_02 ** 2 - 2.49219496 * log10_Z_02 ** 3 -
                    0.63848738 * log10_Z_02 ** 4)
        Tout_o = (17.84778000 - 7.45345690 * log10_Z_02 - 48.9606685 *
                  log10_Z_02 ** 2 - 40.05386135 * log10_Z_02 ** 3 -
                  9.09331816 * log10_Z_02 ** 4)
        Tout_pi = (0.00022582 - 0.00186899 * log10_Z_02 + 0.00388783 *
                   log10_Z_02 ** 2 + 0.00142402 * log10_Z_02 ** 3 -
                   0.00007671 * log10_Z_02 ** 4)
        # caculate Rstar in units of Rsolar
        Rstar = ((Tout_theta * Mstar_Tout ** 2.5 + Tout_l *
                  Mstar_Tout ** 6.5 +
                  Tout_kpa * Mstar_Tout ** 11 + Tout_lbda *
                  Mstar_Tout ** 19 +
                  Tout_mu * Mstar_Tout ** 19.5) /
                 (Tout_nu + Tout_eps * Mstar_Tout ** 2 + Tout_o *
                  Mstar_Tout ** 8.5 + Mstar_Tout ** 18.5 + Tout_pi *
                  Mstar_Tout ** 19.5))

        dmdt = (dmdt * np.sqrt(Mhbase / self._Mh) *
                (self._Mstar / Mstarbase) ** 2.0 * (Rstarbase / Rstar) ** 1.5)
        # tpeak ~ Mh^(1/2) * Mstar^(-1)
        time = (time * np.sqrt(self._Mh / Mhbase) * (Mstarbase / self._Mstar) *
                (Rstar / Rstarbase) ** 1.5)

        time = time / DAY_CGS  # time is now in days to match self._times
        tfallback = np.copy(time[0])
        self._rest_t_explosion = kwargs['resttexplosion']  # units = days

        # ----------- EXTRAPOLATE dm/dt TO EARLY TIMES -------------
        # use power law to fit : dmdt = b*t^xi

        if self.EXTRAPOLATE and self._rest_t_explosion > self._times[0]:
            dfloor = min(dmdt)  # will be at late times if using James's
            # simulaiton data (which already has been late time extrap.)

            # not within 1% of floor, extrapolate --> NECESSARY?
            if dmdt[0] >= dfloor * 1.01:

                # try shifting time before extrapolation to make power law drop
                # off more suddenly around tfallback
                time = time + 0.9 * tfallback
                # this will ensure extrapolation will extend back to first
                # transient time.
                # requires self._rest_t_explosion > self._times[0]
                # time = (time - tfallback + self._rest_t_explosion -
                #        self._times[0])

                ipeak = np.argmax(dmdt)  # index of peak

                # the following makes sure there is enough prepeak sampling for
                # good extrapolation
                if ipeak < 1000:
                    prepeakfunc = interp1d(time[:ipeak], dmdt[:ipeak])
                    prepeaktimes = np.logspace(np.log10(time[0]),
                                               np.log10(time[ipeak - 1]), 1000)
                    # prepeaktimes = np.linspace(time[0], time[ipeak - 1],
                    #                           num=1000)
                    if prepeaktimes[-1] > time[ipeak - 1]:
                        prepeaktimes[-1] = time[ipeak - 1]
                    if prepeaktimes[0] < time[0]:
                        prepeaktimes[0] = time[0]
                    prepeakdmdt = prepeakfunc(prepeaktimes)
                else:
                    prepeaktimes = time[:ipeak]
                    prepeakdmdt = dmdt[:ipeak]

                start = 0

                # last index of first part of data used to get power law fit
                index1 = int(len(prepeakdmdt) * 0.1)
                # last index of second part of data used to get power law fit
                index2 = int(len(prepeakdmdt) * 0.15)

                t1 = prepeaktimes[start:index1]
                d1 = prepeakdmdt[start:index1]

                t2 = prepeaktimes[index2 - (index1 - start):index2]
                d2 = prepeakdmdt[index2 - (index1 - start):index2]

                # exponent for power law fit
                xi = np.log(d1 / d2) / np.log(t1 / t2)
                xiavg = np.mean(xi)

                # multiplicative factor for power law fit
                b1 = d1 / (t1 ** xiavg)

                bavg = np.mean(b1)

                tfloor = 0.01 + 0.9 * tfallback  # want first time ~0 (0.01)

                indexext = len(time[time < prepeaktimes[index1]])

                textp = np.linspace(tfloor, time[int(indexext)], num=ipeak * 5)
                dextp = bavg * (textp ** xiavg)

                time = np.concatenate((textp, time[int(indexext) + 1:]))

                time = time - 0.9 * tfallback  # shift back to original times

                dmdt = np.concatenate((dextp, dmdt[int(indexext) + 1:]))

        # try aligning first fallback time of simulation
        # (whatever first time is before early t extrapolation)
        # with parameter texplosion

        time = time - tfallback + self._rest_t_explosion

        tpeak = time[np.argmax(dmdt)]

        timeinterpfunc = interp1d(time, dmdt)

        lengthpretimes = len(np.where(self._times < time[0])[0])
        lengthposttimes = len(np.where(self._times > time[-1])[0])

        # this removes all extrapolation by interp1d by setting dmdtnew = 0
        # outside bounds of self._times
        dmdt1 = np.zeros(lengthpretimes)
        dmdt3 = np.zeros(lengthposttimes)
        # include len(self._times) instead of just using -lengthposttimes
        # for indexing in case lengthposttimes == 0
        dmdt2 = timeinterpfunc(self._times[lengthpretimes:(len(self._times) -
                                                           lengthposttimes)])
        dmdtnew = np.append(dmdt1, dmdt2)
        dmdtnew = np.append(dmdtnew, dmdt3)

        dmdtnew[dmdtnew < 0] = 0  # set floor for dmdt

        self._efficiency = kwargs['efficiency']
        # luminosities in erg/s
        luminosities = (self._efficiency * dmdtnew *
                        c.c.cgs.value * c.c.cgs.value)
        # -------------- EDDINGTON LUMINOSITY CUT -------------------
        # Assume solar metallicity for now

        # 0.2*(1 + X) = mean Thomson opacity
        kappa_t = 0.2 * (1 + 0.74)
        Ledd = (FOUR_PI * c.G.cgs.value * self._Mh * M_SUN_CGS *
                C_CGS / kappa_t)

        # 2 options for soft Ledd cuts, try both & see what fits stuff better
        # luminosities = np.where(
        #    luminosities > Ledd, (1. + np.log10(luminosities/Ledd)) * Ledd,
        #    luminosities)
        luminosities = (luminosities * Ledd / (luminosities + Ledd))

        return {'dense_luminosities': luminosities, 'Rstar': Rstar,
                'tpeak': tpeak, 'beta': self._beta, 'starmass': self._Mstar,
                'dmdt': dmdtnew, 'Ledd': Ledd}



"""Definitions for the `TDEConstraints` class."""
import astropy.constants as c

from mosfit.constants import C_CGS, M_SUN_CGS
from mosfit.modules.constraints.constraint import Constraint


class TDEConstraints(Constraint):
    """TDE constraints.

    1. rp > rs --> the pericenter radius must be greater than the Schwarzschild
    radius or the bh will swallow the star whole (no disruption or flare)
    """

    def __init__(self, **kwargs):
        """Initialize module."""
        super(TDEConstraints, self).__init__(**kwargs)

    def process(self, **kwargs):
        """Process module. Add constraints below."""
        self._score_modifier = 0.0
        self._rp = kwargs[self.key('rp')]  # already in cgs
        self._bhmass = kwargs['bhmass']
        self._Rs = (2* c.G.cgs.value * self._bhmass * M_SUN_CGS /
                    (C_CGS * C_CGS))

        # Pericenter radius is getting close to Schwarzschild radius
        # if (self._Rg / self._rp > 0.1):
        #       soft limit
        #    self._score_modifier -= 1000.0**(10.0*(self._Rg / self._rp - 0.9))

        self._score_modifier -= 10.0 * (self._Rs / self._rp) ** 3

        return {self.key('score_modifier'): self._score_modifier}



# -*- coding: UTF-8 -*-
"""Definitions for `Nester` class."""

import gc
import pickle
import sys

import numpy as np
from astrocats.catalog.model import MODEL
from astrocats.catalog.quantity import QUANTITY
from mosfit.samplers.sampler import Sampler
from mosfit.utils import pretty_num


class Nester(Sampler):
    """Fit transient events with the provided model."""

    _MAX_ACORC = 5
    _REPLACE_AGE = 20

    def __init__(
        self, fitter, model=None, iterations=2000, burn=None, post_burn=None,
            num_walkers=None, convergence_criteria=None,
            convergence_type='psrf', gibbs=False, fracking=True,
            frack_step=20, **kwargs):
        """Initialize `Nester` class."""
        super(Nester, self).__init__(fitter, num_walkers=num_walkers, **kwargs)

        self._model = model
        self._iterations = iterations
        self._burn = burn
        self._post_burn = post_burn
        self._cc = convergence_criteria
        self._ct = convergence_type
        self._gibbs = gibbs
        self._fracking = fracking
        self._frack_step = frack_step

        self._upload_model = None
        self._ntemps = 1

    def _get_best_kmat(self):
        """Get the kernel matrix associated with best current scoring model."""
        sout = self._model.run_stack(
            self._results.samples[np.unravel_index(
                np.argmax(self._results.logl),
                self._results.logl.shape)],
            root='objective')

        kmat = sout.get('kmat')
        kdiag = sout.get('kdiagonal')
        variance = sout.get('obandvs', sout.get('variance'))
        if kdiag is not None and kmat is not None:
            kmat[np.diag_indices_from(kmat)] += kdiag
        elif kdiag is not None and kmat is None:
            kmat = np.diag(kdiag + variance)

        return kmat

    def append_output(self, modeldict):
        """Append output from the nester to the model description."""
        modeldict[MODEL.SCORE] = {
            QUANTITY.VALUE: pretty_num(self._logz, sig=6),
            QUANTITY.E_VALUE: pretty_num(self._e_logz, sig=6),
            QUANTITY.KIND: 'Log(z)'
        }
        modeldict[MODEL.STEPS] = str(self._niter)

    def prepare_output(self, check_upload_quality, upload):
        """Prepare output for writing to disk and uploading."""
        self._pout = [self._results.samples]
        self._lnprobout = [self._results.logl]
        self._weights = [np.exp(self._results.logwt - max(
            self._results.logwt))]
        tweight = np.sum(self._weights)
        self._weights = [x / tweight for x in self._weights]

        if check_upload_quality:
            pass

    def run(self, walker_data):
        """Use nested sampling to determine posteriors."""
        from dynesty import DynamicNestedSampler
        from dynesty.dynamicsampler import stopping_function, weight_function
        from mosfit.fitter import ln_likelihood, draw_from_icdf

        prt = self._printer

        if len(walker_data):
            prt.message('nester_not_use_walkers', warning=True)

        ndim = self._model._num_free_parameters

        if self._num_walkers:
            self._nwalkers = self._num_walkers
        else:
            self._nwalkers = 2 * ndim

        self._nlive = 20 * ndim

        self._lnprob = None
        self._lnlike = None

        prt.message('nmeas_nfree', [self._model._num_measurements, ndim])

        nested_dlogz_init = self._cc
        post_thresh = self._cc

        max_iter = self._iterations if self._ct is None else np.inf
        if max_iter <= 0:
            return

        s_exception = None
        iter_denom = None if self._ct is not None else self._iterations

        # Save a few things from the dynesty run for diagnostic purposes.
        scales = []

        try:
            sampler = DynamicNestedSampler(
                ln_likelihood, draw_from_icdf, ndim,
                pool=self._pool, sample='rwalk',
                queue_size=max(self._pool.size, 1))
            # Perform initial sample.
            ncall = sampler.ncall
            self._niter = sampler.it - 1
            for li, res in enumerate(sampler.sample_initial(
                dlogz=nested_dlogz_init, nlive=self._nlive
            )):
                ncall0 = ncall
                (worst, ustar, vstar, loglstar, logvol,
                 logwt, self._logz, logzvar, h, nc, worst_it,
                 propidx, propiter, eff, delta_logz) = res

                ncall += nc
                self._niter += 1
                max_iter -= 1

                if max_iter < 0:
                    break

                if (self._fitter._maximum_walltime is not False and
                        self.time_running() >
                        self._fitter._maximum_walltime):
                    prt.message('exceeded_walltime', warning=True)
                    break

                self._results = sampler.results

                scales.append(sampler.results.scale)

                kmat = self._get_best_kmat()
                # The above added 1 call.
                ncall += 1

                self._e_logz = np.sqrt(logzvar)
                prt.status(
                    self, 'baseline', kmat=kmat,
                    iterations=[self._niter, iter_denom],
                    nc=ncall - ncall0, ncall=ncall, eff=eff,
                    logz=[self._logz, self._e_logz,
                          delta_logz, nested_dlogz_init],
                    loglstar=[loglstar],
                    time_running=self.time_running(),
                    maximum_walltime=self._fitter._maximum_walltime)

            if max_iter >= 0:
                prt.status(
                    self, 'starting_batches', kmat=kmat,
                    iterations=[self._niter, iter_denom],
                    nc=ncall - ncall0, ncall=ncall, eff=eff,
                    logz=[self._logz, self._e_logz,
                          delta_logz, nested_dlogz_init],
                    loglstar=[loglstar],
                    time_running=self.time_running(),
                    maximum_walltime=self._fitter._maximum_walltime)

            n = 0
            while max_iter >= 0:
                n += 1
                if (self._fitter._maximum_walltime is not False and
                        self.time_running() >
                        self._fitter._maximum_walltime):
                    prt.message('exceeded_walltime', warning=True)
                    break

                self._results = sampler.results

                scales.append(sampler.results.scale)

                stop, stop_vals = stopping_function(
                    self._results, return_vals=True, args={
                        'post_thresh': post_thresh})
                stop_post, stop_evid, stop_val = stop_vals
                if not stop:
                    logl_bounds = weight_function(self._results)
                    self._logz, self._e_logz = self._results.logz[
                        -1], self._results.logzerr[-1]
                    for res in sampler.sample_batch(
                            logl_bounds=logl_bounds,
                            nlive_new=int(np.ceil(self._nlive / 2))):
                        (worst, ustar, vstar, loglstar, nc,
                         worst_it, propidx, propiter, eff) = res
                        ncall0 = ncall

                        ncall += nc
                        self._niter += 1
                        max_iter -= 1

                        self._results = sampler.results

                        kmat = self._get_best_kmat()
                        # The above added 1 call.
                        ncall += 1

                        prt.status(
                            self, 'batching', kmat=kmat,
                            iterations=[self._niter, iter_denom],
                            batch=n, nc=ncall - ncall0, ncall=ncall, eff=eff,
                            logz=[self._logz, self._e_logz], loglstar=[
                                logl_bounds[0], loglstar,
                                logl_bounds[1]], stop=stop_val,
                            time_running=self.time_running(),
                            maximum_walltime=self._fitter._maximum_walltime)

                        if max_iter < 0:
                            break
                    sampler.combine_runs()
                else:
                    break

                # self._results.summary()
                # prt.nester_status(self, desc='sampling')

        except (KeyboardInterrupt, SystemExit):
            prt.message('ctrl_c', error=True, prefix=False, color='!r')
            s_exception = sys.exc_info()
        except Exception:
            print('Scale history:')
            print(scales)
            pickle.dump(sampler.results, open(
                self._fitter._event_name + '-dynesty.pickle', 'wb'))
            self._pool.close()
            raise

        if max_iter < 0:
            prt.message('max_iter')

        if s_exception is not None:
            self._pool.close()
            if (not prt.prompt('mc_interrupted')):
                sys.exit()

        sampler.reset()
        gc.collect()
